#coding: utf-8

# https://github.com/GCQ2119216031/DDA-SKF/tree/dcad0b455f2d436bafe03b03ce07394f54f075e4
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler

import numpy as np
import pandas as pd
import os
from subprocess import call

import calendar
import time
current_GMT = time.gmtime()

## /!\ Only tested on Linux
class DDA_SKF(BasicModel):
    def __init__(self, params=None):
        try:
            call("octave -v", shell=True)
        except:
            raise ValueError("Please install Octave.")
        try:
            call("octave --silent --eval \"pkg load statistics;\"", shell=True)
        except:
            raise ValueError("Please install Octave package statistics.")
        params = params if (params is not None) else self.default_parameters()
        super(DDA_SKF, self).__init__(params)
        self.scalerS, self.scalerP = None, None
        self.name = "DDA_SKF" 
        self.estimator = None
        self.DDA_SKF_filepath = None

    def default_parameters(self):
        params = {
            "beta" : 0.4, "lamuda" : 2**(-16), 
            "sep_feature": "-",
        }
        return params

    def preprocessing(self, dataset, is_training=True, inf=2):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        S_ = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        P_ = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        if (all([self.sep_feature in str(f) for f in dataset.item_features])):
            types_feature = [str(f).split(self.sep_feature)[0] for f in dataset.item_features]
            S_lst = [S_[:,np.argwhere(np.array(types_feature)==tf)].T for tf in list(set(types_feature))]
            S_lst = [x.reshape(x.shape[1:]) for x in S_lst]
            S_lst = [x if (x.shape[0]==x.shape[1]) else np.corrcoef(x.T) for x in S_lst]
        else:
            S_lst = [S_.T] if (S_.shape[0]==S_.shape[1]) else [np.corrcoef(S_)]
        if (all([self.sep_feature in str(f) for f in dataset.user_features])):
            types_feature = [str(f).split(self.sep_feature)[0] for f in dataset.user_features]
            P_lst = [P_[:,np.argwhere(np.array(types_feature)==tf)].T for tf in list(set(types_feature))]
            P_lst = [x.reshape(x.shape[1:]) for x in P_lst]
            P_lst = [x if (x.shape[0]==x.shape[1]) else np.corrcoef(x.T) for x in P_lst]
        else:
            P_lst = [P_.T] if (P_.shape[0]==P_.shape[1]) else [np.corrcoef(P_)]
        Y = dataset.ratings.toarray().copy()
        Y[Y==-1] = 0
        keep_ids_dr = (np.sum(Y,axis=1)!=0)
        keep_ids_di = (np.sum(Y,axis=0)!=0)
        P_lst = [P[keep_ids_di,:][:,keep_ids_di] for P in P_lst]
        P_lst = [P for P in P_lst if (not np.isnan(np.mean(P)))]
        S_lst = [S[keep_ids_dr,:][:,keep_ids_dr] for S in S_lst]
        S_lst = [S for S in S_lst if (not np.isnan(np.mean(S)))]
        Y_ = Y[keep_ids_dr,:][:,keep_ids_di]
        return [S_lst, P_lst, Y_, keep_ids_dr, keep_ids_di] if (is_training) else [Y]
        
    ## https://raw.githubusercontent.com/GCQ2119216031/DDA-SKF/dcad0b455f2d436bafe03b03ce07394f54f075e4/src/Novel_drug_prediction.m
    def model_fit(self, S_lst, P_lst, Y, keep_ids_dr, keep_ids_di):
        time_stamp = calendar.timegm(current_GMT)
        filefolder = "DDA_SKF_%s" % time_stamp 
        call("mkdir -p %s/" % filefolder, shell=True)
        call("wget -qO %s/DDA_SKF.m 'https://raw.githubusercontent.com/GCQ2119216031/DDA-SKF/dcad0b455f2d436bafe03b03ce07394f54f075e4/src/Novel_drug_prediction.m'" % filefolder, shell=True)
        call("sed -i '1,56d' %s/DDA_SKF.m" % filefolder, shell=True)
        np.savetxt("%s/Y.csv" % filefolder, Y, delimiter=",")
        for i, S in enumerate(S_lst): 
            np.savetxt("%s/S_%d.csv" % (filefolder, i+1), S, delimiter=",")
        for i, P in enumerate(P_lst): 
            np.savetxt("%s/P_%d.csv" % (filefolder, i+1), P, delimiter=",")
        cmd = ["pkg load statistics; interaction_matrix = csvread('Y.csv')'; [dn,dr] = size(interaction_matrix); train_interaction_matrix = interaction_matrix"]
        cmd1 = ["K1 = []"]+["K1(:,:,%d)=csvread('P_%d.csv')" % (i+1, i+1) for i, _ in enumerate(P_lst)]+["K1(:,:,%d)=interaction_similarity(train_interaction_matrix, '1')" % (len(P_lst)+1)]
        cmd2 = ["K2 = []"]+["K2(:,:,%d)=csvread('S_%d.csv')" % (i+1, i+1) for i, _ in enumerate(S_lst)]+["K2(:,:,%d)=interaction_similarity(train_interaction_matrix, '2')" % (len(S_lst)+1)]
        cmd += cmd1+cmd2
        cmd += ["K_COM1=SKF({"+",".join(["K1(:,:,%d)" % (i+1) for i,_ in enumerate(cmd1[1:])])+"},%d,10,0.2)" % min(12, P.shape[0]-1)]
        cmd += ["K_COM2=SKF({"+",".join(["K2(:,:,%d)" % (i+1) for i,_ in enumerate(cmd1[1:])])+"},%d,6,0.4)" % min(20, S.shape[0]-1)] 
        cmd += ["score_matrix = LapRLS(K_COM1,K_COM2,train_interaction_matrix,%f,%f)" % (self.lamuda, self.beta)]
        cmd += ["csvwrite('score_matrix.csv', real(score_matrix))"]
        cmd = (";\n".join(cmd))+";"
        call("echo \"function DDA_SKF\n%s\nend\n\n\" | cat - %s/DDA_SKF.m > %s/DDA_SKF2.m" % (cmd,filefolder,filefolder), shell=True)
        call("mv %s/DDA_SKF2.m %s/DDA_SKF.m" % (filefolder,filefolder), shell=True)
        call("sed -i 's/squaredeuclidean/sqeuclidean/g' %s/DDA_SKF.m" % filefolder, shell=True)
        call("cd %s/ && octave 'DDA_SKF.m'" % (filefolder), shell=True)
        preds_ = np.loadtxt("%s/score_matrix.csv" % filefolder, delimiter=",").T
        preds = np.zeros((len(keep_ids_dr), len(keep_ids_di)))
        for ii, i in enumerate(np.argwhere(keep_ids_dr)):
            preds[i,keep_ids_di] = preds_[ii,:]
        self.estimator = {
            "predictions" : preds,
        }
        call("rm -rf %s/" % filefolder, shell=True)

    def model_predict_proba(self, Y):
        assert Y.shape==self.estimator["predictions"].shape
        return self.estimator["predictions"]


#coding: utf-8

## https://github.com/bbjy/PSGCN/tree/e9edea02d76e3593fe678c8cabf82dc6aaa3a65a
from stanscofi.models import BasicModel
from torch_geometric.data import DataLoader
from benchscofi.implementations import PSGCNImplementation
import numpy as np

class PSGCN(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(PSGCN, self).__init__(params)
        assert self.preprocessing_str in ["Perlman_procedure", "meanimputation_standardize", "same_feature_preprocessing"]
        self.model = None
        self.scalerS, self.scalerP, self.filter = None, None, None
        self.name = "PSGCN"

    def default_parameters(self):
        params = {
            "seed": 1234,
            "hop": 2, # number of neighbor
            "reg_lambda": 0.002,
            "k": 30,
            "lr": 1e-3, # learning rat
            "latent_dim": [64, 64, 1],
            "epochs": 2,#30, # number of epochs
            "batch_size": 128, # batch size during training
            "dropout": 0.4, # random drops neural node and edge with this prob
            "force_undirected": False, # in edge dropout, force (x, y) and (y, x) to be dropped together
            "valid_interval": 1, 
            "device": "cpu", #"cuda"
            "preprocessing_str": "meanimputation_standardize", "subset": None,
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        split_data_dict = PSGCNImplementation.load_k_fold(dataset.ratings.toarray(), self.seed)
        graphs = PSGCNImplementation.extract_subgraph(split_data_dict, self, k=0)
        if (is_training or self.model is None):
            self.model = PSGCNImplementation.PSGCN(graphs, latent_dim=self.latent_dim, k=self.k, dropout=self.dropout,
                force_undirected=self.force_undirected)
        return [graphs] if (is_training) else [graphs, dataset.folds.data.shape[0]]

    def model_fit(self, train_graphs):
        PSGCNImplementation.train_multiple_epochs(train_graphs, train_graphs, self.model, self)
        self.model.eval()

    def model_predict_proba(self, test_graphs, n):
        test_loader = DataLoader(test_graphs, n, shuffle=False, num_workers=2)
        outs = []
        current_n = 0
        for data in test_loader:       
            y_true = data.y.view(-1).cpu().detach().numpy()
            outs.append(self.model(data).cpu().detach().numpy())
            current_n += outs[-1].shape[0]
            #if (current_n>=n):
            break
        outs = np.concatenate(tuple(outs), axis=0)
        return outs


#coding:utf-8

from scipy.optimize import minimize, minimize_scalar
from scipy.stats import norm
from scipy.spatial.distance import cdist
from scipy.special import xlogy
from sklearn.metrics import roc_curve as ROC
import matplotlib.pyplot as plt

import pandas as pd
import numpy as np

def generate_Censoring_dataset(pi=0.3,c=0.3,N=100,nfeatures=50,mean=0.5,std=1,exact=True,random_state=1234):
    '''
    Generate a synthetic dataset under the censoring setting. Assume that $s \in \{-1,1\}$ are the true labels, $y \in \{0,1\}$ are the accessible labels (note that accessible negative samples are missing), and $v \in \mathbb{R}^d$ are the feature vectors. Samples $(v,s) \sim p(v,s)$, and then are made accessible as follows $y \sim p(\cdot \mid v, s=1)$ and $\mathbb{P}(y \/neq 0 \mid v, s=-1) = p(y=-1 \mid v, s=1) = 0$.

    Reference: Elkan, Charles, and Keith Noto. "Learning classifiers from only positive and unlabeled data." Proceedings of the 14th ACM SIGKDD international conference on Knowledge discovery and data mining. 2008.

    ...

    Parameters
    ----------
    pi : float
        Prob(s=1). Comprised in ]0,1[
    c : float
        Prob(s=1 | y!=0). Comprised in ]0,1[
    N : int
        Total number of (item, user) pairs
    nfeatures : int
        Total number of features. Should be a multiple of 2
    mean : float
        Mean (resp., -mean) of the distribution of positive (resp., negative) samples
    std : float
        Standard deviation of the distributions of positive and negative samples
    exact : bool
        If set to True, the proportions in pi and c are exactly respected (e.g., pi=#positive/#total) instead of drawn under a Bernouilli law of mean pi (or c)
    random_state : int
        Random seed

    Returns
    ----------
    dataset_di : dict
        Argument to feed to function stanscofi.datasets.Dataset to create a dataset
    labels_mat : pandas DataFrame
        Corresponding matrix of true labels (i.e., without unlabeled samples)
    '''
    assert nfeatures%2==0
    assert pi>0 and pi<1
    assert c>0 and c<1
    np.random.seed(random_state)
    ## Generate feature matrices for unlabeled samples
    Nsqrt = int(np.sqrt(N)+1)
    if (exact):
        NPos, NNeg = int(pi*Nsqrt), Nsqrt-int(pi*Nsqrt)
    else:
        NN = np.random.binomial(1, pi, size=Nsqrt)
        NPos, NNeg = np.sum(NN), np.sum(NN==0)
    assert NPos+NNeg==Nsqrt
    ### All user feature vectors
    users = np.random.normal(0,std,size=(nfeatures//2,Nsqrt))
    ### All positive pairs
    PosItems = np.random.normal(mean,std,size=(nfeatures//2,NPos))
    ### All negative pairs
    NegItems = np.random.normal(-mean,std,size=(nfeatures//2,NNeg))
    ### All item feature vectors
    items = np.concatenate((PosItems, NegItems), axis=1)
    ### True label matrix
    labels_mat = np.asarray(np.zeros((Nsqrt,Nsqrt)), dtype=int)
    labels_mat[:NPos,:] = 1
    labels_mat[NPos:,:] = -1
    ## Generate accessible ratings = y among positive samples with probability c
    if (exact):
        ids_ls = list(range(Nsqrt*NPos))
        np.random.shuffle(ids_ls)
        NlabPos = np.asarray(np.zeros(Nsqrt*NPos), dtype=int)
        NlabPos[ids_ls[:int(c*Nsqrt*NPos)]] = 1
    else:
        NlabPos = np.random.binomial(1, c, size=Nsqrt*NPos)
    ratings_mat = np.copy(labels_mat)
    ratings_mat[:NPos,:] *= NlabPos.reshape((NPos, Nsqrt)) ## hide some of the positive
    ratings_mat[NPos:,:] = 0 ## hide all negative
    ## Input to stanscofi
    user_list, item_list, feature_list = [list(map(str,x)) for x in [range(Nsqrt), range(Nsqrt), range(nfeatures//2)]]
    ratings_mat = pd.DataFrame(ratings_mat, columns=user_list, index=item_list).astype(int)
    labels_mat = pd.DataFrame(labels_mat, columns=user_list, index=item_list).astype(int)
    users = pd.DataFrame(users, index=feature_list, columns=user_list)
    items = pd.DataFrame(items, index=feature_list, columns=item_list)
    return {"ratings": ratings_mat, "users": users, "items": items}, labels_mat

def generate_CaseControl_dataset(N=100,nfeatures=50,pi=0.3,sparsity=0.01,imbalance=0.03,mean=0.5,std=1,exact=True,random_state=1234):
    '''
    Generate a synthetic dataset under the case-control setting. Assume that $s \in \{-1,1\}$ are the true labels, $y \in \{-1,0,1\}$ are the accessible labels, and $v \in \mathbb{R}^d$ are the feature vectors. Positive pairs $v \sim p_+ = p(\cdot | y=+1)$, negative pairs $v \sim p_- = p(\cdot | y=-1)$, and unlabeled pairs $v \sim p_u = \pi p_+ + (1-\pi)p_-$ (where $\pi := \mathbb{P}(s = 1) \in (0,1)$ is the class-prior probability).

    Reference: Kato, Masahiro, Takeshi Teshima, and Junya Honda. "Learning from positive and unlabeled data with a selection bias." International conference on learning representations. 2018.

    ...

    Parameters
    ----------
    N : int
        Total number of (item, user) pairs
    nfeatures : int
        Total number of features. Should be a multiple of 2
    pi : float
        Prob(s=1). Comprised in ]0,1[
    sparsity : float
        Percentage of labeled samples. Comprised in ]0,1[
    imbalance : float
        Ratio of the numbers of negative and positive labeled samples
    mean : float
        Mean (resp., -mean) of the distribution of positive (resp., negative) samples
    std : float
        Standard deviation of the distributions of positive and negative samples
    exact : bool
        If set to True, the proportions in pi and c are exactly respected (e.g., pi=#positive/#total) instead of drawn under a Bernouilli law of mean pi (or c)
    random_state : int
        Random seed

    Returns
    ----------
    dataset_di : dict
        Argument to feed to function stanscofi.datasets.Dataset to create a dataset
    labels_mat : pandas DataFrame
        Corresponding matrix of true labels (i.e., without unlabeled samples)
    '''
    assert nfeatures%2==0
    assert pi>0 and pi<1
    assert sparsity>0 and sparsity<1
    np.random.seed(random_state)
    ## Generate feature matrices for unlabeled samples (from positive dist with probability pi)
    Nsqrt = int(np.sqrt(N))
    if (exact):
        NPos = int(pi*np.sqrt(N))
        NNeg = Nsqrt-NPos
    else:
        NIsPos = np.random.binomial(1, pi, size=Nsqrt)
        NPos, NNeg = np.sum(NIsPos), np.sum(NIsPos==0)
    ### All user feature vectors
    users = np.random.normal(0,std,size=(nfeatures//2,Nsqrt))
    ## Concatenated item feature vectors for positive and negative pairs
    PosItems = np.random.normal(mean,std,size=(nfeatures//2,NPos))
    NegItems = np.random.normal(-mean,std,size=(nfeatures//2,NNeg))
    items = np.concatenate((PosItems, NegItems), axis=1)
    ### True label matrix
    labels_mat = np.asarray(np.zeros((Nsqrt,Nsqrt)), dtype=int)
    labels_mat[:NPos,:] = 1
    labels_mat[NPos:,:] = -1
    ## Generate accessible ratings = y
    ratings_mat = np.copy(labels_mat)*0
    Ni = sparsity/(pi*(1+imbalance))
    Nip = (sparsity-pi*Ni)/(1-pi)
    NNegLab = int(NNeg*Nsqrt*Nip)
    NPosLab = int(NPos*Nsqrt*Ni)
    ids_user_ls = list(range(NPos*Nsqrt))
    np.random.shuffle(ids_user_ls)
    select_pos = np.asarray(np.zeros(NPos*Nsqrt), dtype=int)
    select_pos[ids_user_ls[:NPosLab]] = 1
    select_pos = select_pos.reshape((NPos, Nsqrt))
    ratings_mat[:NPos,:] = select_pos
    ids_user_ls = list(range(NNeg*Nsqrt))
    np.random.shuffle(ids_user_ls)
    select_neg = np.asarray(np.zeros(NNeg*Nsqrt), dtype=int)
    select_neg[ids_user_ls[:NNegLab]] = -1
    select_neg = select_neg.reshape((NNeg, Nsqrt))
    ratings_mat[NPos:,:] = select_neg
    ## Input to stanscofi
    user_list, item_list, feature_list = [list(map(str,x)) for x in [range(Nsqrt), range(Nsqrt), range(nfeatures//2)]]
    ratings_mat = pd.DataFrame(ratings_mat, columns=user_list, index=item_list).astype(int)
    labels_mat = pd.DataFrame(labels_mat, columns=user_list, index=item_list).astype(int)
    users = pd.DataFrame(users, index=feature_list, columns=user_list)
    items = pd.DataFrame(items, index=feature_list, columns=item_list)
    return {"ratings": ratings_mat, "users": users, "items": items}, labels_mat

def data_aided_estimation(scores_all, true_all, estimator_type=[1,2,3][0]): 
    '''
    Class prior estimation under the case-control setting, that is, 

    Assume that $s \in \{-1,1\}$ are the true labels, $y \in \{0,1\}$ are the accessible labels (note that accessible negative samples are missing), and $v \in \mathbb{R}^d$ are the feature vectors. Samples $(v,s) \sim p(v,s)$, and then are made accessible as follows $y \sim p(\cdot \mid v, s=1)$ and $\mathbb{P}(y \/neq 0 \mid v, s=-1) = p(y=-1 \mid v, s=1) = 0$.

    Three estimators $e_1$, $e_2$ and $e_3$ of $c := \mathbb{P}(s=1 \mid y \/neq 0)$. Given a trained classifier $\widehat{\/theta}$, and a validation set $\mathcal{V} := \{ (v,y) \mid y \in \{-1,0,1\}, v \in \mathbb{R}^d \}$,
    $$ e_1 := \/frac{1}{|\{v \mid (v,+1) \in \mathcal{V}\}|}\sum_{(v,+1) \in \mathcal{V}} (f_{\widehat{\/theta}}(v))_+\;;  e_2 := \/frac{\sum_{(v',+1) \in \mathcal{V}} (f_{\widehat{\/theta}}(v'))_+}{\sum_{(v,y) \in \mathcal{V}} (f_{\widehat{\/theta}}(v))_+}\;; e_3 := \max_{(v,y) \in \mathcal{V}} (f_{\widehat{\/theta}}(v))_+\;. $$

    One can retrieve an approximation of $\pi:=\mathbb{P}(s=1)$ by using $c\pi = \mathbb{P}(y=1) \/approx \sum_{(v',+1) \in \mathcal{V}} (f_{\widehat{\/theta}}(v'))_+$
$$\hat{\pi}_i := \/frac{e_i^{-1}}{|\mathcal{V}|}\sum_{(v,+1) \in \mathcal{V}} (f_{\widehat{\/theta}}(v))_+\;.$$

    Reference: Charles Elkan and Keith Noto. Learning classifiers from only positive and unlabeled data. In Proceedings of the 14th ACM SIGKDD international conference on Knowledge discovery and data mining, pages 213–220, 2008.
    '''
    assert estimator_type in [1,2,3]
    assert (scores_all>=0).all() and (scores_all<=1).all()
    assert scores_all.shape[0] == true_all.shape[0]
    sum_pos = (true_all>0).astype(int).dot(scores_all)
    size = len(scores_all)
    if (estimator_type == 1):
        est_c = sum_pos/np.sum(true_all>0)
    elif (estimator_type==2):
        est_c = sum_pos/np.sum(scores_all)
    else:
        est_c = np.max(scores_all) ## in the paper (but mean is used in pulearn)
    est_pi = sum_pos/(est_c*size)
    return est_c, est_pi

def roc_aided_estimation(scores_all, true_all, regression_type=[1,2][0], show_plot=False, verbose=False):
    '''
    Class prior estimation under the case-control setting, that is, 

    Assume that $s \in \{-1,1\}$ are the true labels, $y \in \{-1,0,1\}$ are the accessible labels, and $v \in \mathbb{R}^d$ are the feature vectors. Positive pairs $v \sim p_+ = p(\cdot | y=+1)$, negative pairs $v \sim p_- = p(\cdot | y=-1)$, and unlabeled pairs $v \sim p_u = \pi p_+ + (1-\pi)p_-$ (where $\pi := \mathbb{P}(s = 1) \in (0,1)$ is the class-prior probability).

    [1, Theorem $4$] shows that if the supports for $p_+$ and $p_-$ are different
$$\hat{\pi} = -\lim_{\substack{\/alpha \/rightarrow 1\/\/ \/alpha < 1}}\/frac{\partial}{\partial \/alpha}\inf_{\/theta \in \Theta} \left\{  \/underbrace{\mathcal{R}_\/text{0-1}(\/theta)}_\/text{Bayes regret} \mid \mathbb{E}_{v \sim p_-}\ell_{0-1}(C_\/theta(v),-1) \leq \/alpha \/right\}(\/alpha)\;.$$

    As mentioned in [2], a possible approach to approximate $\hat{\pi}$ is to regress a specific model (given in [2]) on the points of the corresponding ROC curve, and use the fitted model to extract the slope at the right-hand side of the curve, which is $\hat{\pi}$.

    [1] Scott, Clayton, and Gilles Blanchard. "Novelty detection: Unlabeled data definitely help." Artificial intelligence and statistics. PMLR, 2009.

    [2] Sanderson, Tyler, and Clayton Scott. "Class proportion estimation with application to multiclass anomaly rejection." Artificial Intelligence and Statistics. PMLR, 2014. (arxiv:1306.5056)
    '''
    assert regression_type in [1,2]
    assert scores_all.shape[0] == true_all.shape[0]
    fpr, tpr, _ = ROC(true_all, scores_all)
    if (show_plot):
        plt.plot(fpr, tpr, "b-")
        plt.plot(fpr, fpr, "k--")
        plt.title("ROC curve")
        plt.show()
        plt.close()
    base_fpr = np.linspace(0.001, 0.999, 101) ## alpha false positive rate
    mean_tprs = np.interp(base_fpr, fpr, tpr)
    mean_tprs[0] = 0.0
    ## Empirical (average-user) ROC curve X=base_fpr, Y=mean_tprs
    ## Fit empirical ROC curve onto regression models from C. Lloyd. Regression models for convex ROC curves. Biometrics, 56(3):862–867, September 2000.
    def binomial_deviance(x):
        log = lambda m : xlogy(np.sign(m), m)
        import warnings
        warnings.simplefilter("ignore") #invalid value in power
        power = lambda m, p : np.power(m,p) #np.sign(m)*(np.abs(m))**p
        #def power(m,p):
        #    try:
        #        import warnings
        #        warnings.simplefilter("error")
        #        m[m<0] = 0
        #        return np.power(m,p)
        #    except:
        #        print(m)
        #        print(p)
        #        raise ValueError
        if (regression_type==1):
            gamma, Delta = x.tolist()
            def f(alpha):
                Phi = norm(loc=0.0, scale=1.0)
                Q, inv_Q = np.vectorize(Phi.cdf), np.vectorize(Phi.ppf)
                val = (1-gamma)*Q(inv_Q(alpha)+Delta)+gamma*alpha
                return val
        else:
            gamma, Delta, mu = x.tolist()
            def f(alpha):
                val = (1-gamma)*power(1+Delta*(1/power(alpha,mu)-1), -1/mu)+gamma*alpha
                return val
        dev = -2*np.sum( np.multiply(mean_tprs, log(f(base_fpr))) + np.multiply(1-mean_tprs, log(1-f(base_fpr))) )
        #print((regression_type, x, dev))
        return dev
    x0 = np.array([1]*(regression_type+1)) ## gamma, Delta(, mu if regression_type=2)
    bnds = tuple([(0, 1)]+[(None, None)]*regression_type)
    res = minimize(binomial_deviance, x0, bounds=bnds, method='nelder-mead', options={'xatol': 1e-8, 'disp': verbose, "maxiter": 1000})
    args = res.x.tolist()
    if (regression_type == 1):
        return args[0]
    return (1-args[0])*args[1]+args[0]

def divergence_aided_estimation(X, y, lmb=1, sigma=1., divergence_type=["L1-distance","Pearson"][0], show_plot=False):
    '''
    Class prior estimation under the case-control setting, that is, 

    Assume that $s \in \{-1,1\}$ are the true labels, $y \in \{-1,0,1\}$ are the accessible labels, and $v \in \mathbb{R}^d$ are the feature vectors. Positive pairs $v \sim p_+ = p(\cdot | y=+1)$, negative pairs $v \sim p_- = p(\cdot | y=-1)$, and unlabeled pairs $v \sim p_u = \pi p_+ + (1-\pi)p_-$ (where $\pi := \mathbb{P}(s = 1) \in (0,1)$ is the class-prior probability).

$\lambda$ and $\sigma$ are regularization parameters, and $p$ (resp., $u$) is the total number of positive (resp., unlabeled) samples).

    Using L1-distance penalized divergence [1] amounts to minimizing the following scalar function:

    $$\hat{\pi}_\/text{L1} := \/arg\min_{\pi \in (0,1)} \/frac{1}{\lambda}\sum_{l \leq p+u} ((\/beta_l(\pi))_+)^2-\pi+1 \/text{ and } \/beta_l(\pi) := \/frac{\pi}{u}\sum_{i \leq u} \mathcal{N}(x_l, \sigma^2 \/text{Id})(x_i)-\/frac{1}{p}\sum_{j \leq p} \mathcal{N}(x_l, \sigma^2 \/text{Id})(x_j)\;.$$

    [1] Christoffel, Marthinus, Gang Niu, and Masashi Sugiyama. "Class-prior estimation for learning from positive and unlabeled data." Asian Conference on Machine Learning. PMLR, 2016. (https://proceedings.mlr.press/v45/Christoffel15.pdf)

    Using the Pearson penalized divergence [2] amounts to minimizing the following scalar function:

    $$\hat{\pi}_\/text{Pearson} := \/arg\min_{\pi \in (0,1)} -\/frac{1}{2}\left[^{1-\pi}_{\pi}\/right] H^\/top(G + \lambda R)^{-1}G(G+\lambda R)^{-1}H\left[^{1-\pi}_{\pi}\/right]^\/top+\left[^{1-\pi}_{\pi}\/right] H^\/top (G+\lambda R)^{-1} H\left[^{1-\pi}_{\pi}\/right]^\/top-\/frac{1}{2} $$

    $\/text{ and } H := \left[\/frac{1}{u}\sum_{j \leq u}\left(\mathcal{N}(x_l, \sigma^2 \/text{Id})(x_j)\/right)_{0 \leq l \leq u+p}, \/frac{1}{p}\sum_{i \leq p}\left(\mathcal{N}(x_l, \sigma^2 \/text{Id})(x_i)\/right)_{0 \leq l \leq u+p} \/right] \in \mathbb{R}^{(u+p+1) \/times 2} \;, R := \left[^{0}_{(0)_{(u+p) \/times 1}} ,^{(0)_{1 \/times (u+p)}}_{Id_{(u+p) \/times (u+p)}}\/right] \in \mathbb{R}^{(u+p+1) \/times (u+p+1)} \;,$

    $G:=\/frac{1}{u+p} \sum_{i \leq u+p} \left(\mathcal{N}(x_l, \sigma^2 \/text{Id})(x_i)\/right)_{0 \leq l \leq u+p}^\/top\left(\mathcal{N}(x_l, \sigma^2 \/text{Id})(x_i)\/right)_{0 \leq l \leq u+p} \in \mathbb{R}^{(u+p+1) \/times (u+p+1)}$ where $\/forall x, \mathcal{N}(x_0, \sigma^2 \/text{Id})(x)=1$.

    [2] Du Plessis, Marthinus Christoffel, and Masashi Sugiyama. "Semi-supervised learning of class balance under class-prior change by distribution matching." Neural Networks 50 (2014): 110-119. (arxiv:1206.4677)
    '''
    assert divergence_type in ["L1-distance", "Pearson"]
    pos_x = (y==1).astype(int)
    unl_x = (y<1).astype(int) # (y==0).astype(int)
    basis_mat = np.exp(-cdist(X,X,metric='euclidean')/(2*sigma**2))
    if (divergence_type=="L1-distance"):
        def approx_div(pi):
            betas = [pi/np.sum(pos_x)*basis_mat[l,pos_x].sum() if (pos_x.sum()>0) else 0 for l in range(basis_mat.shape[0])]
            betas = [betas[l]-(1/np.sum(unl_x)*basis_mat[l,unl_x].sum() if (unl_x.sum()>0) else 0) for l in range(basis_mat.shape[0])]
            return (1/lmb)*np.sum([max(0,b)*b for b in betas])-pi+1
    else:
        R1 = np.zeros((1,basis_mat.shape[0]))
        R3 = np.eye(basis_mat.shape[0])
        R = np.concatenate((np.column_stack((0,R1)), np.column_stack((R1.T, R3))), axis=0)
        H = np.array([1/np.sum(x)*np.array([1]+[np.sum([basis_mat[l,i] for i in range(basis_mat.shape[0]) if (x[i])]) for l in range(basis_mat.shape[0])]).T for x in [unl_x, pos_x]]).T
        basis_mat = np.concatenate((np.ones((1,basis_mat.shape[1])), basis_mat), axis=0)
        basis_mat = np.concatenate((np.ones((1,basis_mat.shape[0])).T, basis_mat), axis=1)
        G = (1/basis_mat.shape[0])*np.sum([basis_mat[l,:].reshape((1,-1)).T.dot(basis_mat[l,:].reshape((1,-1))) for l in range(basis_mat.shape[0])], axis=1)
        def approx_div(pi):
            theta = np.array([1-pi, pi]).T
            GlR = np.linalg.pinv(G+lmb*R)
            return -0.5*theta.dot(H.T).dot(GlR).dot(G).dot(GlR).dot(H.dot(theta.T))+theta.dot(H.T).dot(GlR).dot(H.dot(theta.T))-0.5
    if (show_plot):
        plt.figure(figsize=(2,2))
        pi_ls = [0.1*p for p in range(0,10)]
        plt.plot(pi_ls, [approx_div(p) for p in pi_ls], "b-")
        plt.title("penL1(pi) curve")
        plt.show()
        plt.close()
    res = minimize_scalar(approx_div, bounds=(0, 1), method='bounded')
    return res.x



#coding: utf-8

## http://bioinformatics.csu.edu.cn/resources/softs/DrugRepositioning/DRRS/index.html
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler

import numpy as np
import pandas as pd
import os
from subprocess import call

import calendar
import time
current_GMT = time.gmtime()

## /!\ Only tested on Linux
class DRRS(BasicModel):
    def __init__(self, params=None):
        self.MCR_HOME="/usr/local/MATLAB/MATLAB_Compiler_Runtime"
        self.DRRS_path = "http://bioinformatics.csu.edu.cn/resources/softs/DrugRepositioning/DRRS/soft/"
        if (not os.path.exists(self.MCR_HOME)):
            raise ValueError("Please install MATLAB.")
        params = params if (params is not None) else self.default_parameters()
        super(DRRS, self).__init__(params)
        assert self.use_linux
        self.scalerS, self.scalerP = None, None
        self.name = "DRRS" 
        self.predictions = None
        self.DRRS_filepath = "DRRS_L" if (self.use_linux) else "DRRS_W.exe"

    def default_parameters(self):
        params = {
            "use_linux": True, #False: use windows
        }
        return params

    def preprocessing(self, dataset, is_training=True, inf=2):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        S_ = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        X_s = S_ if (S_.shape[0]==S_.shape[1]) else np.corrcoef(S_)
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        P_ = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        X_p = P_ if (P_.shape[0]==P_.shape[1]) else np.corrcoef(P_)
        A_sp = dataset.ratings.toarray().T # users x items
        return [X_s, X_p, A_sp]
        
    def model_fit(self, X_s, X_p, A_sp):
        time_stamp = calendar.timegm(current_GMT)
        filefolder = "DRRS_%s" % time_stamp 
        call("mkdir -p %s/" % filefolder, shell=True) 
        call("wget -qO "+filefolder+"/"+self.DRRS_filepath+" "+self.DRRS_path+self.DRRS_filepath+" && chmod +x "+filefolder+"/"+self.DRRS_filepath, shell=True)      
        drugsim, diseasesim, didra = [x+".txt" for x in ["DrugSim","DiseaseSim","DiDrA"]]
        for x, tx in zip([X_s,X_p,A_sp],[drugsim, diseasesim, didra]):
            pd.DataFrame(x, index=range(x.shape[0]), columns=range(x.shape[1])).to_csv(filefolder+"/"+tx,sep="\t",header=None,index=None)
        os.environ['LD_LIBRARY_PATH'] = "%s/v80/runtime/glnxa64:%s/v80/bin/glnxa64:%s/v80/sys/java/jre/glnxa64/jre/lib/amd64/server:%s/v80/sys/os/glnxa64:%s/v80/sys/java/jre/glnxa64/jre/lib/amd64:%s/v80/sys/java/jre/glnxa64/jre/lib/amd64/native_threads" % tuple([self.MCR_HOME]*6)
        os.environ['XAPPLRESDIR'] = "%s/v80/X11/app-defaults" % self.MCR_HOME
        call(" ".join(["cd", "%s/" % filefolder, "&&", "./"+self.DRRS_filepath, drugsim, diseasesim, didra]), shell=True)
        assert os.path.exists("%s/Result_dr_Mat.txt" % filefolder)
        self.predictions = np.loadtxt("%s/Result_dr_Mat.txt" % filefolder, delimiter="\t").T
        call("rm -rf %s/ %s" % (filefolder, self.DRRS_filepath), shell=True)

    def model_predict_proba(self, X_s, X_p, A_sp):
        return self.predictions


#coding: utf-8

from stanscofi.models import BasicModel
from stanscofi.training_testing import random_simple_split

import os
from subprocess import Popen, call
import numpy as np

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse import csr_matrix
from sklearn.metrics import ndcg_score, roc_auc_score
from collections import Counter
from tqdm import tqdm
import sys

## https://github.com/jilljenn/libmf/tree/9f0d6af350ba342ee52bf28111cbe30335f3baab
def execute_ndcg(trainfile, testfile, outfiles):
    df_train = pd.read_csv(trainfile, sep=' ', names=('user', 'item', 'v'))
    df = pd.read_csv(testfile, sep=' ', names=('user', 'item', 'v'))
    df_test = df
    df['user_id'] = np.unique(df['user'], return_inverse=True)[1]
    df['item_id'] = np.unique(df['item'], return_inverse=True)[1]
    encode_user = dict(df[['user', 'user_id']].drop_duplicates().to_numpy())
    encode_item = dict(df[['item', 'item_id']].drop_duplicates().to_numpy())
    df_train['user_id'] = df_train['user'].map(encode_user)
    df_train['item_id'] = df_train['item'].map(encode_item)

    n_users = df['user'].nunique()
    n_items = df['item'].nunique()
    n_entries, _ = df.shape
    #print(n_users, df['user_id'].max())
    #print(n_items, df['item_id'].max())
    truth = csr_matrix((df['v'] * n_entries, (df['user_id'], df['item_id'])), shape=(n_users, n_items))

    for filename in outfiles:
        out = pd.read_csv(filename, names=('pred',))
        #print(filename, out.shape, df['user_id'].shape)
        sparse = csr_matrix((out['pred'], (df['user_id'], df['item_id'])), shape=(n_users, n_items))

        #print('global AUC', roc_auc_score(df['v'], out['pred']))

        df['pred'] = out['pred']

        ndcg_values = []
        ndcg10_values = []
        auc_values = []
        better_auc_values = []
        normalized_row_auc_values = []
        for user_id in tqdm(df['user_id'].unique()):
            this_user = df_test.query("user_id == @user_id")
            #print(this_user)
            this_user_pos = this_user.query("v == 1")
            nb_pos = len(this_user_pos)
            this_user_neg = this_user.query("v == 0")
            nb_neg = len(this_user_neg)
            if (nb_pos*nb_neg==0):
                ndcg_values.append(0)
                ndcg10_values.append(0)
                auc_values.append(0)
                better_auc_values.append(0)
                normalized_row_auc_values.append(0)
                continue ## ignore
            this_user_pairs = this_user_pos.merge(this_user_neg, how='cross')
            #print(this_user_pairs.head())
            okay_pairs = this_user_pairs.apply(lambda row: row['pred_x'] > row['pred_y'], axis=1)
            #print(okay_pairs)
            normalized_row_auc_values.append(sum(okay_pairs) / (nb_pos * nb_neg))
            # print(this_user_pairs.shape)
            # print(normalized_row_auc_values)

            # nb_neg = this_user['v'].value_counts().loc[0]
            # nb_pos = this_user['v'].value_counts().loc[1]
            better_auc_values.append(roc_auc_score([max(0,s) for s in this_user['v']], [max(0,s) for s in this_user['pred']]))
            # break

            #print(user_id, '=', nb_pos, nb_neg, nb_pos + nb_neg, sum(okay_pairs) / (nb_pos * nb_neg), better_auc_values[-1])

            test_set = list(set(range(n_items)) - set(df_train.query("user_id == @user_id")['item_id'].tolist()))
            # print(n_items - len(test_set))
            user_truth = truth[user_id, test_set].toarray()
            # print(Counter(user_truth.reshape(-1).tolist()))
            user_pred = sparse[user_id, test_set].toarray()
            user_truth[user_truth<0] = 0
            user_pred[user_pred<0] = 0
            ndcg_values.append(ndcg_score(user_truth, user_pred))
            ndcg10_values.append(ndcg_score(user_truth, user_pred, k=10))

            try:
                auc_values.append(roc_auc_score(user_truth.reshape(-1),
                                        user_pred.reshape(-1)))
            except ValueError as e:
                assert e.args[0]=="Only one class present in y_true. ROC AUC score is not defined in that case." ## ignore users with a single class
                #print(e) 
                #print(df.query("user_id == @user_id").shape)
                #print(df.query("user_id == @user_id and v == 1"))
                #print(df_train.query("user_id == @user_id"))
                #print(user_id, len(test_set), Counter(user_truth.tolist()[0]))
                ndcg_values.append(0)
                ndcg10_values.append(0)
                auc_values.append(0)
                better_auc_values.append(0)
                normalized_row_auc_values.append(0)
                continue ## ignore
                #break
    print(filename)
    print('ndcg =', np.mean(ndcg_values))
    print('ndcg@10 =', np.mean(ndcg10_values))
    print('row auc =', np.mean(auc_values))
    print('row-oriented auc =', np.mean(better_auc_values))
    print('normalized row-oriented auc =', np.mean(normalized_row_auc_values))
    return np.mean(ndcg_values), np.mean(ndcg10_values), np.mean(auc_values), np.mean(better_auc_values), np.mean(normalized_row_auc_values)

class LibMFWrapper(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(LibMFWrapper, self).__init__(params)
        self.libmf_folder = "libmf/"
        self.params = params
        if (not os.path.exists(self.libmf_folder)):
            Popen("git clone -q https://github.com/jilljenn/libmf.git".split(" "))
        if (not os.path.exists(self.libmf_folder+"mf-train") or not os.path.exists(self.libmf_folder+"mf-predict")):
            call(("cd "+self.libmf_folder+" && make"), shell=True)
        self.name = "LibMFWrapper"

    def default_parameters(self):
        params = {
            'f': 12, # loss function: squared error (L2-norm)
            'l2': 0.01, # L2-regularization parameter
            'k': 32, # number of dimensions
            "a": 0.001, # coefficient of negative entries' loss
            "s": 10, # number of threads
            "c": 0.0001, # value of negative entries
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        if (not is_training): ## positive and negative (0, -1) pairs
            rats = ( dataset.ratings.toarray()[dataset.folds.toarray()>0] ).ravel()
            mat = np.column_stack((dataset.folds.row, dataset.folds.col, rats))
            #mat = mat[mat[:,-1]>0]
            return [mat, rats]
        folds, _ = random_simple_split(dataset, 0.1, metric="euclidean")
        mat_lst = []
        for fds in folds: ## only positive pairs
            sb_dt = dataset.subset(fds)
            rats = ( sb_dt.ratings.toarray()[sb_dt.folds.toarray()>0] ).ravel()
            mat = np.column_stack((sb_dt.folds.row, sb_dt.folds.col, rats))
            mat = mat[mat[:,-1]>0]
            mat_lst.append(mat)
        return mat_lst

    def model_fit(self, mat_tr, mat_te, rm=True):
        np.savetxt(self.libmf_folder+"mat.tr.txt", mat_tr, fmt='%d')
        np.savetxt(self.libmf_folder+"mat.te.txt", mat_te, fmt='%d')
        conv_args = [(x,self.params[x],"d" if (" 'int'>" in str(type(self.params[x]))) else "f") for x in self.params]
        cmd = self.libmf_folder+"mf-train "+" ".join(["-"+a+" "+(("%"+f) % v) for a,v,f in conv_args])
        cmd += " -p %s %s %s" % (self.libmf_folder+"mat.te.txt", self.libmf_folder+"mat.tr.txt", self.libmf_folder+"ocmf_model.txt")
        process = Popen(cmd.split(" "))
        process.wait()
        with open(self.libmf_folder+"ocmf_model.txt", "r") as f:
            self.model = f.read()
        if (rm):
            process = Popen(("rm -f "+self.libmf_folder+"mat.*.txt "+self.libmf_folder+"ocmf_model.txt").split(" "))
            process.wait()

    def model_predict_proba(self, mat, rats, ev=12, rm=True):
        np.savetxt(self.libmf_folder+"mat.txt", mat, fmt='%d')
        with open(self.libmf_folder+"ocmf_model.txt", "w") as f:
            f.write(self.model)
        cmd = self.libmf_folder+"mf-predict -e %d %s %s %s" % (ev, self.libmf_folder+"mat.txt", 
            self.libmf_folder+"ocmf_model.txt", self.libmf_folder+"ocmf_output.txt"
        )
        process = Popen(cmd.split(" "))
        process.wait()
        #preds = np.zeros(rats.shape[0])
        #preds[rats==1] = np.loadtxt(self.libmf_folder+"ocmf_output.txt")
        preds = np.loadtxt(self.libmf_folder+"ocmf_output.txt")
        if (rm):
            process = Popen(("rm -f "+self.libmf_folder+"mat.txt "+self.libmf_folder+"ocmf_*.txt").split(" "))
            process.wait()
        return preds


#coding: utf-8

## https://github.com/bioinfomaticsCSU/MBiRW/tree/d0487b2a43e37a7ee4026959cb052e2527611fde
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler

import numpy as np
import pandas as pd
import os
from subprocess import call
import scipy.io
import warnings

import calendar
import time
current_GMT = time.gmtime()

## /!\ Only tested on Linux
class MBiRW(BasicModel):
    def __init__(self, params=None):
        try:
            call("octave -v", shell=True)
        except:
            raise ValueError("Please install Octave.")
        params = params if (params is not None) else self.default_parameters()
        super(MBiRW, self).__init__(params)
        self.scalerS, self.scalerP = None, None
        self.name = "MBiRW" 
        self.predictions = None
        self.MBiRW_filepath = None

    def default_parameters(self):
        params = {
            "alpha": 0.3, "l":2, "r":2, "d": np.log(9999),
        }
        return params

    def preprocessing(self, dataset, is_training=True, inf=2):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        S_ = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        X_s = S_ if (S_.shape[0]==S_.shape[1]) else np.corrcoef(S_)
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        P_ = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        X_p = P_ if (P_.shape[0]==P_.shape[1]) else np.corrcoef(P_)
        A_sp = dataset.ratings.toarray().T # users x items
        return [X_s, X_p, A_sp, np.array(dataset.item_list), np.array(dataset.user_list)] if (is_training) else [X_s, X_p, A_sp]
        
    def model_fit(self, X_s, X_p, A_sp, item_list, user_list):
        time_stamp = calendar.timegm(current_GMT)
        filefolder = "MBiRW_%s" % time_stamp 
        call("mkdir -p %s/" % filefolder, shell=True)
        repo_url = "https://raw.githubusercontent.com/bioinfomaticsCSU/MBiRW/d0487b2a43e37a7ee4026959cb052e2527611fde"
        call("wget -qO %s/normFun.m '%s/Code/normFun.m'" % (filefolder, repo_url), shell=True)
        call("wget -qO %s/setparFun.m '%s/Code/setparFun.m'" % (filefolder, repo_url), shell=True)
        call("wget -qO - '%s/Code/nManiCluester.m' | sed '5,55d' > %s/nManiCluester.m" % (repo_url, filefolder), shell=True)
        call("wget -qO %s/cluster_one-1.0.jar '%s/Code/cluster_one-1.0.jar'" % (filefolder, repo_url), shell=True)
        np.savetxt("%s/X_p.csv" % filefolder, X_p, delimiter=",")
        np.savetxt("%s/A_sp.csv" % filefolder, A_sp, delimiter=",")
        np.savetxt("%s/X_s.csv" % filefolder, X_s, delimiter=",")
        with open("%s/s_names.csv" % filefolder, "w") as f:
            f.write(",".join(item_list))
        with open("%s/p_names.csv" % filefolder, "w") as f:
            f.write(",".join(user_list))
        ## https://github.com/bioinfomaticsCSU/MBiRW/blob/d0487b2a43e37a7ee4026959cb052e2527611fde/Code/MBiRW.m
        newWrr, newWdd = np.copy(X_s), np.copy(X_p)
        dr, dn = newWrr.shape[1], newWdd.shape[1]
        with open("%s/DrugsP.txt" % filefolder, "w") as f:
            lines = ["\t".join([str(item_list[i]), str(item_list[j]), str(newWrr[i,j])]) for i in range(dr) for j in range(i) if (newWrr[i,j]>0)]
            f.write("\n".join(lines))
        with open("%s/DiseasesP.txt" % filefolder, "w") as f:
            lines = ["\t".join([str(user_list[i]), str(user_list[j]), str(newWdd[i,j])]) for i in range(dn) for j in range(i) if (newWdd[i,j]>0)]
            f.write("\n".join(lines))
        call('cd %s/ && java -jar "cluster_one-1.0.jar"  "DrugsP.txt" -F csv > DrugsC.txt' % filefolder, shell=True)
        call('cd %s/ && java -jar "cluster_one-1.0.jar"  "DiseasesP.txt" -F csv > DiseasesC.txt' % filefolder, shell=True)
        cmd = "Wdd = csvread('X_p.csv');Wdr = csvread('A_sp.csv');Wrr = csvread('X_s.csv');Wrname = csvread('s_names.csv');Wdname = csvread('p_names.csv');Wrd = Wdr';A = Wrd;alpha=%f;l=%d;r=%d;d=%f;dn = size(Wdd,1);dr = size(Wrr,1);newWrr = csvread('X_s.csv');newWdd = csvread('X_p.csv');cr = setparFun(Wrd,Wrr);cd = setparFun(Wdr,Wdd);LWrr = 1./(1+exp(cr*Wrr+d));LWdd = 1./(1+exp(cd*Wdd+d));[RWrr,RWdd] = nManiCluester(LWrr,LWdd,newWrr,newWdd,Wrname,Wdname);normWrr = normFun(RWrr);normWdd = normFun(RWdd);R0 = A/sum(A(:));Rt = R0;for t=1:max(l,r);ftl = 0;ftr = 0;if(t<=l);nRtleft = alpha * normWrr*Rt + (1-alpha)*R0;ftl = 1;end;if(t<=r);nRtright = alpha * Rt * normWdd + (1-alpha)*R0;ftr = 1;end;Rt =  (ftl*nRtleft + ftr*nRtright)/(ftl + ftr);end;csvwrite('Rt.csv', Rt);" % (self.alpha, self.l, self.r, self.d)
        call("cd %s/ && octave --silent --eval \"%s\"" % (filefolder, cmd), shell=True)
        self.predictions = np.loadtxt("%s/Rt.csv" % filefolder, delimiter=",")
        call("rm -rf %s/" % filefolder, shell=True)

    def model_predict_proba(self, X_s, X_p, A_sp):
        return self.predictions


#coding: utf-8

## https://github.com/BioinformaticsCSU/BNNR/tree/68f8e98c02459189b6eeac68a86306ccc1da0374
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler

import numpy as np
import pandas as pd
import os
from subprocess import call

import calendar
import time
current_GMT = time.gmtime()

## /!\ Only tested on Linux
class BNNR(BasicModel):
    def __init__(self, params=None):
        try:
            call("octave -v", shell=True)
        except:
            raise ValueError("Please install Octave.")
        params = params if (params is not None) else self.default_parameters()
        super(BNNR, self).__init__(params)
        self.scalerS, self.scalerP = None, None
        self.name = "BNNR" 
        self.estimator = None
        self.BNNR_filepath = None

    def default_parameters(self):
        params = {
            "maxiter": 300,
            "alpha": 1,
            "beta": 10,
            "tol1": 2e-3,
            "tol2": 1e-5,
        }
        return params

    def preprocessing(self, dataset, is_training=True, inf=2):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        S_ = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        S_ = np.nan_to_num(S_, nan=0.0) ##
        X_s = S_ if (S_.shape[0]==S_.shape[1]) else np.corrcoef(S_)
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        P_ = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        P_ = np.nan_to_num(P_, nan=0.0) ##
        X_p = P_ if (P_.shape[0]==P_.shape[1]) else np.corrcoef(P_)
        A_sp = dataset.ratings.toarray().T # users x items
        return [X_s, X_p, A_sp]
        
    def model_fit(self, X_s, X_p, A_sp):
        time_stamp = calendar.timegm(current_GMT)
        filefolder = "BNNR_%s" % time_stamp 
        call("mkdir -p %s/" % filefolder, shell=True)
        call("wget -qO %s/BNNR.m 'https://raw.githubusercontent.com/BioinformaticsCSU/BNNR/master/BNNR.m'" % filefolder, shell=True)
        call("wget -qO %s/svt.m 'https://raw.githubusercontent.com/BioinformaticsCSU/BNNR/master/svt.m'" % filefolder, shell=True)
        np.savetxt("%s/X_p.csv" % filefolder, X_p, delimiter=",")
        np.savetxt("%s/A_sp.csv" % filefolder, A_sp, delimiter=",")
        np.savetxt("%s/X_s.csv" % filefolder, X_s, delimiter=",")
        cmd = "Wdd = csvread('X_p.csv');Wdr = csvread('A_sp.csv');Wrr = csvread('X_s.csv');T = [Wrr, Wdr'; Wdr, Wdd];[WW,iter] = BNNR(%d, %d, T, double(T ~= 0), %f, %f, %d, 0, 1);[t1, t2] = size(T);[dn,dr] = size(Wdr);M_recovery = WW((t1-dn+1) : t1, 1 : dr);csvwrite('M_recovery.csv', M_recovery);csvwrite('iter.csv', iter);" % (self.alpha, self.beta, self.tol1, self.tol2, self.maxiter)
        call("cd %s/ && octave --silent --eval \"%s\"" % (filefolder, cmd), shell=True)
        self.estimator = {
            "niter" : int(np.loadtxt("%s/iter.csv" % filefolder, delimiter=",")),
            "predictions" : np.loadtxt("%s/M_recovery.csv" % filefolder, delimiter=",").T,
        }
        call("rm -rf %s/" % filefolder, shell=True)

    def model_predict_proba(self, X_s, X_p, A_sp):
        return self.estimator["predictions"]


#coding: utf-8

from stanscofi.models import BasicModel
from stanscofi.preprocessing import preprocessing_XY

import numpy as np
import os
import random

import tensorflow as tf
from tensorflow.python.keras.models import Sequential, Model
from tensorflow.python.keras.layers import Dense, Input

class SimpleNeuralNetwork(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(SimpleNeuralNetwork, self).__init__(params)
        self.name = "SimpleNeuralNetwork"
        self.scalerP, self.scalerS, self.filter = None, None, None
        assert self.preprocessing_str in ["Perlman_procedure", "meanimputation_standardize", "same_feature_preprocessing"]
        assert len(self.layers_dims)>0

    def default_parameters(self):
        params = {
            "layers_dims": [16,32], "decision_threshold": 0.5, "preprocessing_str": "meanimputation_standardize", 
            "subset": None, "steps_per_epoch":1, "epochs":50, "random_state": 1234,
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        X, y, scalerS, scalerP, filter_ = preprocessing_XY(dataset, self.preprocessing_str, subset_=self.subset, filter_=self.filter, scalerS=self.scalerS, scalerP=self.scalerP, inf=2, njobs=1)
        X = np.nan_to_num(X, nan=0.)
        self.filter = filter_
        self.scalerS = scalerS
        self.scalerP = scalerP
        return [X, y] if (is_training) else [X]
        
    def model_fit(self, X, y):
        os.environ['PYTHONHASHSEED']=str(self.random_state)
        os.environ['TF_CUDNN_DETERMINISTIC'] = '1'  # new flag present in tf 2.0+
        os.environ['TF_DETERMINISTIC_OPS'] = '1'
        tf.random.set_seed(self.random_state)
        random.seed(self.random_state)
        np.random.seed(self.random_state)
        tf.config.threading.set_inter_op_parallelism_threads(1)
        tf.config.threading.set_intra_op_parallelism_threads(1)
        log_ratio = Sequential(
            [Dense(self.layers_dims[0], input_dim=X.shape[1], activation='relu')]
            +[Dense(x, activation="relu") for x in self.layers_dims[1:]]
            +[Dense(1)]
        )
        x_p = Input(shape=(X.shape[1],))
        x_q = Input(shape=(X.shape[1],))
        log_ratio_p = log_ratio(x_p)
        log_ratio_q = log_ratio(x_p)
        self.model, XX, YY = self.nn_creation(x_p, x_q, log_ratio_p, log_ratio_q, X, y)
        hist = self.model.fit(x=XX, y=YY, steps_per_epoch=self.steps_per_epoch, epochs=self.epochs, verbose=0)

    def model_predict_proba(self, X):
        raise NotImplemented

    def nn_creation(self, x_p, x_q, log_ratio_p, log_ratio_q, X, y):
        raise NotImplemented

class SimpleBinaryClassifier(SimpleNeuralNetwork):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(SimpleBinaryClassifier, self).__init__(params)
        self.name = "SimpleBinaryClassifier"

    def _binary_crossentropy(self, log_ratio_p, log_ratio_q):
        loss_q, loss_p = [
            tf.nn.sigmoid_cross_entropy_with_logits(logits=x,labels=tf.ones_like(x) if (ix>0) else tf.zeros_like(x))
            for ix, x in enumerate([log_ratio_q, log_ratio_p])
        ]
        return tf.reduce_mean(loss_p + loss_q)

    def nn_creation(self, x_p, x_q, log_ratio_p, log_ratio_q, X, y):
        model = Model(inputs=[x_p, x_q], outputs=[log_ratio_p, log_ratio_q])
        model.add_loss(self._binary_crossentropy(log_ratio_p, log_ratio_q))
        model.compile(optimizer='rmsprop', loss=None, metrics=['accuracy'])
        XX = [X[y==1,:], X[y<1]]
        YY = [tf.ones(np.sum(y==1)), tf.zeros(np.sum(y<1))]
        if (XX[0].shape[0]>XX[1].shape[0]):
            n = XX[0].shape[0]-XX[1].shape[0]
            XX[1] = np.concatenate(( XX[1], np.tile(XX[1][0,:],(n, 1)) ), axis=0)
            YY[1] = tf.zeros(np.sum(y<1)+n)
        elif (XX[0].shape[0]<XX[1].shape[0]):
            n = XX[1].shape[0]-XX[0].shape[0]
            XX[0] = np.concatenate(( XX[0], np.tile(XX[0][0,:],(n, 1)) ), axis=0)
            YY[0] = tf.ones(np.sum(y==1)+n)
        return model, XX, YY

    def model_predict_proba(self, X):
        p_pred, q_pred = self.model.predict(x=[X, X])
        preds = np.concatenate((tf.nn.sigmoid(p_pred).numpy(), tf.nn.sigmoid(q_pred).numpy()), axis=1)
        return preds[:,1]
        



#coding:utf-8

## Concatenated contents of https://github.com/ljatynu/NIMCGCN/tree/a0798ed29ae329dd71bff096ffc678527cc4099e/code
from __future__ import division
from torch import nn, optim


import torch as t
from torch import nn
from torch_geometric.nn import conv

import csv
import torch as t
import random

class Config(object):
    def __init__(self):
        self.data_path = '../data'
        self.validation = 5
        self.save_path = '../data'
        self.epoch = 300
        self.alpha = 0.2


class Myloss(nn.Module):
    def __init__(self, alpha=1.):
        self.alpha = alpha
        super(Myloss, self).__init__()

    def forward(self, one_index, zero_index, target, input):
        loss = nn.MSELoss(reduction='none')
        loss_sum = loss(input, target)
        return (1-self.alpha)*loss_sum[one_index].sum()+self.alpha*loss_sum[zero_index].sum()


class Sizes(object):
    def __init__(self, dataset):
        self.m = dataset['mm']['data'].size(0)
        self.d = dataset['dd']['data'].size(0)
        self.fg = 256
        self.fd = 256
        self.k = 32


def train(model, train_data, optimizer, opt):
    model.train()
    regression_crit = Myloss(model.alpha)
    one_index = train_data[2][0].cuda().t().tolist() if (t.cuda.is_available()) else train_data[2][0].t().tolist()
    zero_index = train_data[2][1].cuda().t().tolist() if (t.cuda.is_available()) else train_data[2][1].t().tolist()

    def train_epoch():
        model.zero_grad()
        score = model(train_data)
        loss = regression_crit(one_index, zero_index, train_data[4].cuda() if (t.cuda.is_available()) else train_data[4], score)
        loss.backward()
        optimizer.step()
        return loss
    for epoch in range(1, model.epoch+1):
        train_reg_loss = train_epoch()
        if (epoch%opt.display_epoch==0):
            print("LOSS %.5f (epoch %d)" % (train_reg_loss.item()/(len(one_index[0])+len(zero_index[0])), epoch))



def main():
    opt = Config()
    dataset = prepare_data(opt)
    sizes = Sizes(dataset)
    train_data = Dataset(opt, dataset)
    for i in range(opt.validation):
        print('-'*50)
        model = Model(sizes)
        #model.cuda()
        optimizer = optim.Adam(model.parameters(), lr=0.001)
        train(model, train_data[i], optimizer, opt)



def read_csv(path):
    with open(path, 'r', newline='') as csv_file:
        reader = csv.reader(csv_file)
        md_data = []
        md_data += [[float(i) for i in row] for row in reader]
        return t.FloatTensor(md_data)


def read_txt(path):
    with open(path, 'r', newline='') as txt_file:
        reader = txt_file.readlines()
        md_data = []
        md_data += [[float(i) for i in row.split()] for row in reader]
        return t.FloatTensor(md_data)


def get_edge_index(matrix):
    edge_index = [[], []]
    for i in range(matrix.size(0)):
        for j in range(matrix.size(1)):
            if matrix[i][j] != 0:
                edge_index[0].append(i)
                edge_index[1].append(j)
    return t.LongTensor(edge_index)


def prepare_data(opt):
    dataset = dict()
    dataset['md_p'] = read_csv(opt.data_path + '\\m-d.csv')
    dataset['md_true'] = read_csv(opt.data_path + '\\m-d.csv')

    zero_index = []
    one_index = []
    for i in range(dataset['md_p'].size(0)):
        for j in range(dataset['md_p'].size(1)):
            if dataset['md_p'][i][j] < 1:
                zero_index.append([i, j])
            if dataset['md_p'][i][j] >= 1:
                one_index.append([i, j])
    random.shuffle(one_index)
    random.shuffle(zero_index)
    zero_tensor = t.LongTensor(zero_index)
    one_tensor = t.LongTensor(one_index)
    dataset['md'] = dict()
    dataset['md']['train'] = [one_tensor, zero_tensor]

    dd_matrix = read_csv(opt.data_path + '\\d-d.csv')
    dd_edge_index = get_edge_index(dd_matrix)
    dataset['dd'] = {'data': dd_matrix, 'edge_index': dd_edge_index}

    mm_matrix = read_csv(opt.data_path + '\\m-m.csv')
    mm_edge_index = get_edge_index(mm_matrix)
    dataset['mm'] = {'data': mm_matrix, 'edge_index': mm_edge_index}
    return dataset



class Dataset(object):
    def __init__(self, opt, dataset):
        self.data_set = dataset
        self.nums = opt.validation

    def __getitem__(self, index):
        return (self.data_set['dd'], self.data_set['mm'],
                self.data_set['md']['train'], None,
                self.data_set['md_p'], self.data_set['md_true'])

    def __len__(self):
        return self.nums




class Model(nn.Module):
    def __init__(self, sizes):
        super(Model, self).__init__()

        self.fg = sizes.fg
        self.fd = sizes.fd
        self.k = sizes.k
        self.m = sizes.m
        self.d = sizes.d
        self.gcn_x1 = conv.GCNConv(self.fg, self.fg)
        self.gcn_y1 = conv.GCNConv(self.fd, self.fd)
        self.gcn_x2 = conv.GCNConv(self.fg, self.fg)
        self.gcn_y2 = conv.GCNConv(self.fd, self.fd)

        self.linear_x_1 = nn.Linear(self.fg, 256)
        self.linear_x_2 = nn.Linear(256, 128)
        self.linear_x_3 = nn.Linear(128, 64)

        self.linear_y_1 = nn.Linear(self.fd, 256)
        self.linear_y_2 = nn.Linear(256, 128)
        self.linear_y_3 = nn.Linear(128, 64)

    def forward(self, input):
        #t.manual_seed(1)
        x_m = t.randn(self.m, self.fg)
        x_d = t.randn(self.d, self.fd)

        ## ensure that weights are nonnegative
        weights1 = t.relu(input[1]['data'][input[1]['edge_index'][0], input[1]['edge_index'][1]].cuda() if (t.cuda.is_available()) else input[1]['data'][input[1]['edge_index'][0], input[1]['edge_index'][1]])
        X1 = t.relu(self.gcn_x1(x_m.cuda() if (t.cuda.is_available()) else x_m
		, input[1]['edge_index'].cuda() if (t.cuda.is_available()) else input[1]['edge_index']
		, weights1
        ))

        ## ensure that weights are nonnegative
        weights2 = t.relu(input[1]['data'][input[1]['edge_index'][0], input[1]['edge_index'][1]].cuda() if (t.cuda.is_available()) else input[1]['data'][input[1]['edge_index'][0], input[1]['edge_index'][1]])
        X = t.relu(self.gcn_x2(X1, input[1]['edge_index'].cuda() if (t.cuda.is_available()) else input[1]['edge_index']
            , weights2
        ))

        ## ensure that weights are nonnegative
        weights3 = t.relu(input[0]['data'][input[0]['edge_index'][0], input[0]['edge_index'][1]].cuda() if (t.cuda.is_available()) else input[0]['data'][input[0]['edge_index'][0], input[0]['edge_index'][1]])
        Y1 = t.relu(self.gcn_y1(x_d.cuda() if (t.cuda.is_available()) else x_d, input[0]['edge_index'].cuda() if (t.cuda.is_available()) else input[0]['edge_index']
            , weights3
        ))
        ## ensure that weights are nonnegative
        weights4 = t.relu(input[0]['data'][input[0]['edge_index'][0], input[0]['edge_index'][1]].cuda() if (t.cuda.is_available()) else input[0]['data'][input[0]['edge_index'][0], input[0]['edge_index'][1]] )
        Y = t.relu(self.gcn_y2(Y1, input[0]['edge_index'].cuda() if (t.cuda.is_available()) else input[0]['edge_index']
            , weights4
        ))

        x1 = t.relu(self.linear_x_1(X))
        x2 = t.relu(self.linear_x_2(x1))
        x = t.relu(self.linear_x_3(x2))

        y1 = t.relu(self.linear_y_1(Y))
        y2 = t.relu(self.linear_y_2(y1))
        y = t.relu(self.linear_y_3(y2))

        return x.mm(y.t())





#coding: utf-8

from stanscofi.models import BasicModel
import numpy as np
import libmf
from libmf import mf
from multiprocessing import cpu_count

class LibMF(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(LibMF, self).__init__(params)
        self.name = "LibMF"
        self.model = mf.MF(params)

    def default_parameters(self):
        params = {
            "fun": 0,
            "k": 8,
            "nr_threads": max(cpu_count()-2,1),
            "nr_bins": 26,
            "nr_iters": 20,
            "lambda_p1": 0.04,
            "lambda_p2": 0.0,
            "lambda_q1": 0.04,
            "lambda_q2": 0.0,
            "eta": 0.1,
            "do_nmf": False,
            "quiet": False,
            "copy_data": True,
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        row, col = dataset.folds.row, dataset.folds.col
        rats = dataset.ratings.toarray()[row, col].ravel()
        return [np.vstack(tuple([row, col, rats])).transpose().copy()] if (is_training) else [np.vstack(tuple([row, col])).transpose().copy()]

    def model_fit(self, data):
        self.model.fit(data)

    def model_predict_proba(self, data):
        return self.model.predict(data)


#coding: utf-8

from stanscofi.models import BasicModel
from stanscofi.preprocessing import preprocessing_XY
import numpy as np

 ## UNLABELED ARE -1 POSITIVE ARE 1
import pulearn

class PulearnWrapper(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(PulearnWrapper, self).__init__(params)
        assert self.classifier in ["ElkanotoPuClassifier","WeightedElkanotoPuClassifier","BaggingPuClassifier"]
        self.scalerS, self.scalerP, self.filter = None, None, None
        self.name = self.classifier
        self.estimator = eval("pulearn."+self.classifier)(**self.classifier_params)

    def default_parameters(self):
        from sklearn.svm import SVC
        params = {
            "classifier_params": {
                "estimator": SVC(C=10, kernel='rbf', gamma=0.4, probability=True),
                "hold_out_ratio": 0.2,
                #"labeled":10, "unlabeled":20, "n_estimators":15,
            },
            "classifier": "ElkanotoPuClassifier",
            "preprocessing_str": "meanimputation_standardize",
            "subset": None,
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        X, y, scalerS, scalerP, filter_ = preprocessing_XY(dataset, self.preprocessing_str, subset_=self.subset, filter_=self.filter, scalerS=self.scalerS, scalerP=self.scalerP, inf=2, njobs=1)
        self.filter = filter_
        self.scalerS = scalerS
        self.scalerP = scalerP
        ## Unlabeled samples are -1 in pulearn
        y[y<1] = -1
        return [X, y] if (is_training) else [X]
        
    def model_fit(self, X, y):
        self.estimator.fit(X, y)

    def model_predict_proba(self, X):
        preds = self.estimator.predict_proba(X)
        return preds



#coding: utf-8

from stanscofi.models import BasicModel
import numpy as np
from scipy.sparse import csr_matrix

from benchscofi.implementations import AlternatingLeastSquares

class ALSWR(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(ALSWR, self).__init__(params)
        self.name = "ALSWR"
        self.model = AlternatingLeastSquares.ALSWR(**{(k if (k!="random_state") else "seed"):params[k] for k in params})

    def default_parameters(self):
        params = AlternatingLeastSquares.alswr_params
        params.update({"random_state": 1354})
        return params

    def preprocessing(self, dataset, is_training=True):
        ## users x items: only 1's and -1's
        if (is_training):
            ratings = csr_matrix((dataset.ratings.data, (dataset.ratings.col, dataset.ratings.row)), shape=dataset.ratings.T.shape)
        else:
            ids = np.argwhere(np.ones(dataset.ratings.shape))
            ratings = csr_matrix((dataset.ratings.toarray().ravel(), (ids[:,1].ravel(), ids[:,0].ravel())), shape=dataset.ratings.T.shape)
        return [ratings]
        
    def model_fit(self, X_train):
        np.random.seed(self.random_state)
        self.model.fit(X_train)

    def model_predict_proba(self, X_test):
        preds = self.model.predict().T
        return preds


#coding: utf-8

from stanscofi.models import BasicModel
import numpy as np
import torch
from torch import nn
from tqdm import tqdm
from scipy.sparse import coo_array
from sklearn.metrics import roc_auc_score, mean_squared_error, average_precision_score

## https://github.com/jilljenn/vae/tree/1d7f09af3bcaebfc5d8fa8cc18033d8bb8ca19bc
from benchscofi.implementations import VariationalInference

class VariationalWrapper(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(VariationalWrapper, self).__init__(params)
        self.name = "VariationalWrapper"
        self.model = None

    def default_parameters(self):
        params = {
            "LEARNING_RATE" : 1,
            "N_VARIATIONAL_SAMPLES" : 1,
            "N_EPOCHS" : 5,
            "DISPLAY_EPOCH_EVERY" : 5,
            "BATCH_SIZE" : 50,
            "EMBEDDING_SIZE" : 3,
            "N_VARIATIONAL_SAMPLES" : 1,
            "optimizer": "Adam",
            "random_state": 1234,
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        N, M = len(dataset.user_list), len(dataset.item_list)
        ## Ratings between 0 and 1: -1 -> 0, 0 -> 0.5, 1 -> 1
        y = dataset.ratings.toarray()[dataset.folds.row, dataset.folds.col].ravel()
        #y[y==0] = 0.5 
        #y[y==-1] = 0.
        X = np.asarray(np.zeros((y.shape[0], N+M)), dtype=np.float64)
        for x in range(y.shape[0]):
            X[x,dataset.folds.col[x]] = 1.
            X[x,dataset.folds.row[x]+N] = 1.
        return [torch.LongTensor(X), torch.LongTensor(y), N, M] if (is_training) else [torch.LongTensor(X)]

    ## source: https://raw.githubusercontent.com/jilljenn/vae/1d7f09af3bcaebfc5d8fa8cc18033d8bb8ca19bc/vfm-torch.py
    def model_fit(self, X, y, N, M):
        torch.manual_seed(self.random_state)
        np.random.seed(self.random_state)
        torch_dataset = torch.utils.data.TensorDataset(X, y)
        nb_samples = len(y)
        train_rmse = train_auc = train_map = 0.
        losses = []
        all_preds = []
        self.BATCH_SIZE = min(self.BATCH_SIZE, nb_samples//2)
        train_iter = torch.utils.data.DataLoader(torch_dataset, batch_size=self.BATCH_SIZE) # , shuffle=True
        self.model = VariationalInference.CF(self.EMBEDDING_SIZE, self.N_VARIATIONAL_SAMPLES, N=N, M=M,
                output="class")
                #output="reg")
        self.optimizer = {
            "Adam": torch.optim.Adam(self.model.parameters(), lr=self.LEARNING_RATE), #weight_decay=1e-4)
            "LBFGS": torch.optim.LBFGS(self.model.parameters(), lr=self.LEARNING_RATE, history_size=10, max_iter=4, line_search_fn='strong_wolfe'),
            "SGD": torch.optim.SGD(self.model.parameters(), lr=self.LEARNING_RATE),
        }[self.optimizer]
        for epoch in tqdm(range(self.N_EPOCHS)):
            losses = []
            pred = []
            truth = []
            for i, (indices, target) in enumerate(train_iter):
                # print('=' * 10, i)
                outputs, _, _, kl_term = self.model(indices)#.squeeze()
                # print(outputs)
                # print('indices', indices.shape, 'target', target.shape, outputs, 'ypred', len(y_pred), 'kl', kl_term.shape)
                # loss = loss_function(outputs, target)
                # print('kl', kl_bias.shape, kl_entity.shape)
                # print(outputs.sample()[:5], target[:5])
                loss = -outputs.log_prob(target.float()).mean() * nb_samples + kl_term
                # print('loss', loss)
                train_auc = -1

                y_pred = outputs.mean.squeeze().detach().numpy().tolist()
                losses.append(loss.item())
                pred.extend(y_pred)
                truth.extend(target)

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                # print('preds', len(y_pred))
                # print('but target', target.shape)
                # print(len(pred), len(truth))
            # optimizer.swap_swa_sgd()

            # End of epoch
            train_auc = roc_auc_score(truth, pred)
            train_map = average_precision_score(truth, pred)

            '''print('test', outputs.sample()[:5], target[:5], loss.item())
                print('variance', torch.sqrt(1 / model.alpha))
                print('bias max abs', model.bias_params.weight.abs().max())
                print('entity max abs', model.entity_params.weight.abs().max())'''

            if epoch % self.DISPLAY_EPOCH_EVERY == 0:
                print('train pred', np.round(pred[:5], 4), truth[:5])
                print(f"Epoch {epoch}: Elbo {np.mean(losses):.4f} " +
                    (f"Minibatch train AUC {train_auc:.4f} " +
                        f"Minibatch train MAP {train_map:.4f}"))

                print('precision', self.model.alpha, 'std dev', torch.sqrt(1 / nn.functional.softplus(self.model.alpha)))
                # print('bias max abs', self.model.bias_params.weight.abs().max())
                # print('entity max abs', self.model.entity_params.weight.abs().max())

    def model_predict_proba(self, X):
        outputs, _, _, _ = self.model(X)
        y_pred = outputs.mean.squeeze().detach().numpy()
        return y_pred


#coding: utf-8

#https://github.com/gu-yaowen/MilGNet/blob/a0a129311a07feb928580ee7695d4321f24fd5c3/baseline/HAN_imp.py
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler
from benchscofi.implementations import HANImplementation

import os
import dgl
import torch
import pandas as pd
import numpy as np
from scipy.sparse import coo_array

class HAN(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(HAN, self).__init__(params)
        self.scalerS, self.scalerP = None, None
        self.device = torch.device('cpu')
        self.scalerS, self.scalerP = None, None
        self.name = "HAN" 
        self.estimator = None

    def default_parameters(self):
        params = {
            "k": 15, "learning_rate": 1e-3, "epoch": 1000, 
            "weight_decay": 0.0, "seed": 42, 
        }
        return params

    def preprocessing(self, dataset, inf=2, is_training=True):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        drug_drug = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        drug_drug = np.nan_to_num(drug_drug, nan=0) ##
        drug_drug = drug_drug if (drug_drug.shape[0]==drug_drug.shape[1]) else np.corrcoef(drug_drug)
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        disease_disease = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        disease_disease = np.nan_to_num(disease_disease, nan=0) ##       
        disease_disease = disease_disease if (disease_disease.shape[0]==disease_disease.shape[1]) else np.corrcoef(disease_disease)
        drug_drug_link = HANImplementation.topk_filtering(drug_drug, self.k)
        disease_disease_link = HANImplementation.topk_filtering(disease_disease, self.k)
        drug_disease = dataset.ratings.toarray()
        drug_disease[drug_disease<0] = 0
        drug_disease_link = np.array(np.where(drug_disease == 1)).T
        disease_drug_link = np.array(np.where(drug_disease.T == 1)).T
        graph_data = {('drug', 'drug-drug', 'drug'): (torch.tensor(drug_drug_link[:, 0]),
                                                  torch.tensor(drug_drug_link[:, 1])),
                  ('drug', 'drug-disease', 'disease'): (torch.tensor(drug_disease_link[:, 0]),
                                                        torch.tensor(drug_disease_link[:, 1])),
                  ('disease', 'disease-drug', 'drug'): (torch.tensor(disease_drug_link[:, 0]),
                                                        torch.tensor(disease_drug_link[:, 1])),
                  ('disease', 'disease-disease', 'disease'): (torch.tensor(disease_disease_link[:, 0]),
                                                              torch.tensor(disease_disease_link[:, 1]))}
        g = dgl.heterograph(graph_data)
        drug_feature = np.hstack((drug_drug, np.zeros(drug_disease.shape)))
        dis_feature = np.hstack((np.zeros(drug_disease.T.shape), disease_disease))
        g.nodes['drug'].data['h'] = torch.from_numpy(drug_feature).to(torch.float32)
        g.nodes['disease'].data['h'] = torch.from_numpy(dis_feature).to(torch.float32)
        data = torch.tensor(np.column_stack((dataset.folds.row, dataset.folds.col)).astype('int64')).to(self.device)
        label = torch.tensor(drug_disease[dataset.folds.row, dataset.folds.col].flatten()).float().to(self.device)
        shp = dataset.ratings.shape
        return [g, data, label, shp] if (is_training) else [g]
        
    def model_fit(self, g_train, train_data, train_label, shp):
        HANImplementation.set_seed(self.seed)
        feature = {'drug': g_train.nodes['drug'].data['h'],
                   'disease': g_train.nodes['disease'].data['h']}

        model = HANImplementation.HAN(in_feats=[feature['drug'].shape[1],
                              feature['disease'].shape[1]],
                    meta_paths=[['drug-disease', 'disease-drug'],
                                ['disease-drug', 'drug-disease'],
                                ['drug-drug', 'drug-disease', 'disease-drug'],
                                ['disease-disease', 'disease-drug', 'drug-disease']])

        model.to(self.device)
        optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate)
        criterion = torch.nn.BCEWithLogitsLoss(pos_weight=torch.tensor(len(torch.where(train_label == 0)[0]) /
                                                                       len(torch.where(train_label == 1)[0])))
        print('BCE loss pos weight: {:.3f}'.format(
            len(torch.where(train_label == 0)[0]) / len(torch.where(train_label == 1)[0])))

        train_idx = torch.tensor([(i.item() - 1) * shp[1] + j
                                  for (i, j) in train_data]).to(self.device)
        for epoch in range(1, self.epoch + 1):
            model.train()

            pred = model(g_train, feature).flatten().flatten()[train_idx]
            pred_score = torch.sigmoid(pred)

            optimizer.zero_grad()
            loss = criterion(pred, train_label)
            loss.backward()
            optimizer.step()

            if epoch % 100 == 0 or epoch == self.epoch - 1:
                model.eval()
                AUC, AUPR = HANImplementation.get_metrics_auc(train_label.detach().cpu().numpy(),
                                            pred_score.detach().cpu().numpy())
                print('Epoch {} Loss: {:.3f}; Train: AUC {:.3f}, '
                      'AUPR {:.3f}'.format(epoch, loss.item(), AUC, AUPR))
        self.estimator = {"predictions": torch.sigmoid(model(g_train, feature)).detach().cpu().numpy()}

    def model_predict_proba(self, g):
        return self.estimator["predictions"]


#coding: utf-8

from stanscofi.models import BasicModel
import numpy as np

from benchscofi.implementations import LogisticMatrixFactorization

class LogisticMF(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(LogisticMF, self).__init__(params)
        self.name = "LogisticMF"
        self.estimator = LogisticMatrixFactorization.LogisticMF(**params)

    def default_parameters(self):
        params = {"counts": np.zeros((5,6)), "num_factors": 2, "reg_param":0.6, "gamma":1.0, "iterations":30}
        return params

    def preprocessing(self, dataset, is_training=True):
        counts = dataset.ratings.toarray().T
        counts[counts<1] = 0
        total = np.sum(counts)
        num_zeros = np.prod(counts.shape)-total
        alpha = num_zeros / total
        #print('alpha %.2f' % alpha)
        counts *= alpha
        return [counts]
        
    def model_fit(self, X):
        self.estimator.counts = X
        self.estimator.num_users = X.shape[0]
        self.estimator.num_items = X.shape[1]
        self.estimator.train_model()

    def model_predict_proba(self, X):
        ## prediction(i,j) = sigmoid(x_i . y_j^T + b_i + b_j)
        fx = self.estimator.user_vectors.dot(self.estimator.item_vectors.T)
        fx += np.tile(self.estimator.user_biases, (1,self.estimator.item_biases.shape[0]))
        fx += np.tile(self.estimator.item_biases, (1,self.estimator.user_biases.shape[0])).T
        preds = np.exp(fx)/(1+np.exp(fx))
        return preds.T








#coding:utf-8

import numpy as np

## https://github.com/cjlin1/libmf/tree/4e14bc7e8bd429abd1bbec2abc7c3926aaa3fb10 
def calc_mpr_auc(scores, dataset, transpose=False): 
    n, m = dataset.ratings.T.shape if (transpose) else dataset.ratings.shape
    nnz = dataset.ratings.data.shape[0]
    y_true_all = dataset.ratings.toarray()[dataset.ratings.row,dataset.ratings.col]
    y_true_all = y_true_all.ravel() if (transpose) else y_true_all.T.ravel()
    scores_all = scores.data.ravel()
    R = np.column_stack((dataset.ratings.row,dataset.ratings.col)).tolist()
    R_dict = {(i,j): y_true_all[ii] for ii, [i,j] in enumerate(R)}
    R = sorted(R)
    sort_by_pred = lambda x, y : x[1]<y[1]
    pos_cnts = [0]*(m+1)
    for i in range(nnz):
        pos_cnts[R[i][int(not transpose)]] += 1
    for i in range(1, m+1):
        pos_cnts[i] += pos_cnts[i-1]
    total_m, total_pos, all_u_mpr, all_u_auc = 0, 0, [], []
    for i in range(m):
        if (pos_cnts[i+1]-pos_cnts[i] < 1):
            continue
        row = [((i,j),R_dict.get((i,j),0)) for j in range(n)]
        pos = 0
        index = [0]*(pos_cnts[i+1]-pos_cnts[i])
        for j in range(pos_cnts[i], pos_cnts[i+1]):
            if (y_true_all[j]<=0):
                continue
            col = R[j][1]
            #row[col].first.r = prob->R[j].r
            index[pos] = col
            pos += 1
        if ((n-pos < 1) or (pos < 1)):
            continue
        total_m += 1
        total_pos += pos
        count = 0
        for k in range(pos):
            tmp = row[count]
            row[count] = row[index[k]]
            row[index[k]] = tmp
            count += 1
        row = [row[i] for i in np.argsort([r for _,r in row])]
        row.reverse()
        u_mpr, u_auc = 0, 0
        for neg_it in range(pos, len(row)):
            if (row[pos-1][1] <= row[neg_it][1]):
                u_mpr += pos
                continue
            left, right = 0, pos-1
            while (left<right):
                mid = int((left+right)/2)
                if (row[mid][1]>row[neg_it][1]):
                    right = mid
                else:
                    left = mid+1
            u_mpr += left
            u_auc += pos-left
        all_u_mpr.append(u_mpr/(n-pos))
        all_u_auc.append(u_auc/(n-pos)/pos)
    return all_u_mpr, all_u_auc, np.sum(all_u_mpr)/total_pos, np.sum(all_u_auc)/total_m

def calc_auc(scores, dataset, transpose=False): 
    user_ids = np.unique(dataset.folds.col) if (transpose) else np.unique(dataset.folds.row)
    n_ignored = 0
    y_true_all = dataset.ratings.toarray()[dataset.folds.row,dataset.folds.col].ravel() 
    scores_all = scores.data.ravel()
    assert y_true_all.shape==scores_all.shape
    rowwise_aucs = []
    for user_id in user_ids:
        user_ids_i = np.argwhere(dataset.folds.col==user_id) if (transpose) else np.argwhere(dataset.folds.row==user_id)
        if (len(user_ids_i)==0):
            n_ignored += 1
            continue
        user_truth = y_true_all[user_ids_i].ravel()
        user_pred = scores_all[user_ids_i].ravel()
        if ((len(np.unique(user_truth))==2) and (1 in user_truth)):
            ## Full
            omegaplus, omegaminus = np.sum(user_truth==1), np.sum(user_truth<1)
            if (omegaplus*omegaminus==0):
                return 0
            num = sum([int(user_pred[j]>user_pred[jp]) for j in np.argwhere(user_truth==1) for jp in np.argwhere(user_truth<1)])
            auc = num/(omegaplus*omegaminus)
            rowwise_aucs.append(auc)
        else:
            rowwise_aucs.append(0)
            n_ignored += 1
    return rowwise_aucs


# coding:utf-8

## concatenation of contents in https://github.com/bbjy/PSGCN/tree/e9edea02d76e3593fe678c8cabf82dc6aaa3a65a

import torch
from torch_geometric.nn import GCNConv

from torch.optim import Adam

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

import math
import torch.nn as nn
import torch.nn.functional as F

from torch.nn import Linear, Conv1d
from torch_geometric.nn import GCNConv, global_sort_pool
from torch_geometric.utils import dropout_adj

import os
import torch
import time
import networkx as nx
import numpy as np
import pandas as pd
import scipy.sparse as ssp
import multiprocessing as mp

from tqdm import tqdm
from scipy import io
from torch_geometric.data import Data, InMemoryDataset, Dataset
from sklearn.model_selection import KFold

import warnings
warnings.simplefilter('ignore', ssp.SparseEfficiencyWarning)
cur_dir = os.path.dirname(os.path.realpath(__file__))
import torch.multiprocessing
torch.multiprocessing.set_sharing_strategy('file_system')

import torch

from torch_geometric.data import DataLoader
from sklearn import metrics

import torch
from torch_geometric.nn import GCNConv

def extract_subgraph(split_data_dict, args, k=0):
    #if args.data_name == 'Gdataset':
    #    print("Using Gdataset with 10% testing...")
    #    (
    #        adj_train, train_labels, train_u_indices, train_v_indices,
    #        test_labels, test_u_indices, test_v_indices
    #    ) = split_data_dict[k]
    #
    #elif args.data_name == 'Cdataset':
    #
    #    print("Using Cdataset with 10% testing...")
    #    (
    #        adj_train, train_labels, train_u_indices, train_v_indices,
    #        test_labels, test_u_indices, test_v_indices
    #    ) = split_data_dict[k]  # load_drug_data(path)
    #
    #else:
    #    print("Using LRSSL with 10% testing...")
    #    (
    #        adj_train, train_labels, train_u_indices, train_v_indices,
    #        test_labels, test_u_indices, test_v_indices
    #    ) = split_data_dict[k]

    (
        adj_train, train_labels, train_u_indices, train_v_indices#,
        #test_labels, test_u_indices, test_v_indices
    ) = split_data_dict[k]

    val_test_appendix = str(k) + '_kfold'
    data_combo = ("a", "a") #(args.data_name, val_test_appendix)

    train_indices = (train_u_indices, train_v_indices)
    #test_indices = (test_u_indices, test_v_indices)

    train_file_path = 'data/{}/{}/train'.format(*data_combo)
    train_graph = MyDataset(train_file_path, adj_train, train_indices, train_labels, args.hop)
    # train_graph = MyDynamicDataset(train_file_path, adj_train, train_indices, train_labels, args.hop)

    #test_file_path = 'data/{}/{}/test'.format(*data_combo)
    #test_graph = MyDataset(test_file_path, adj_train, test_indices, test_labels, args.hop)
    # test_graph = MyDynamicDataset(test_file_path, adj_train, test_indices, test_labels, args.hop)

    return train_graph#, test_graph

class attention_score(torch.nn.Module):
    def __init__(self, in_channels, Conv=GCNConv):
        super(attention_score, self).__init__()
        self.in_channels = in_channels
        self.score_layer = Conv(in_channels, 1)

    def forward(self, x, edge_index, edge_attr=None, batch=None):
        score = self.score_layer(x, edge_index)

        return score


def train_multiple_epochs(train_dataset, test_dataset, model, args):

    train_loader = DataLoader(train_dataset, args.batch_size, shuffle=True, num_workers=2)

    test_size = 1024  # load all test dataset
    test_loader = DataLoader(test_dataset, test_size, shuffle=False,
                             num_workers=2)

    model.to(args.device).reset_parameters()
    optimizer = Adam(model.parameters(), lr=args.lr)

    start_epoch = 1
    pbar = range(start_epoch, args.epochs + start_epoch)

    for epoch in pbar:
        train_loss = train(model, optimizer, train_loader, args.device)

        if epoch % args.valid_interval == 0:
            roc_auc, aupr = evaluate_metric(model, test_loader, args.device)

            print("epoch {}".format(epoch), "train_loss {0:.4f}".format(train_loss),
                  "roc_auc {0:.4f}".format(roc_auc), "aupr {0:.4f}".format(aupr))


def num_graphs(data):
    if data.batch is not None:
        return data.num_graphs
    else:
        return data.x.size(0)


def train(model, optimizer, loader, device):
    model.train()
    total_loss = 0
    pbar = loader
    for data in pbar:
        optimizer.zero_grad()
        true_label = data.to(device)
        predict = model(true_label)
        loss_function = torch.nn.BCEWithLogitsLoss()
        loss = loss_function(predict, true_label.y.view(-1))
        loss.backward()
        total_loss += loss.item() * num_graphs(data)
        optimizer.step()
        torch.cuda.empty_cache()

    return total_loss / len(loader.dataset)


def evaluate_metric(model, loader, device):
    model.eval()
    pbar = loader
    roc_auc, aupr = None, None
    for data in pbar:
        data = data.to(device)
        with torch.no_grad():
            out = model(data)

        y_true = data.y.view(-1).cpu().tolist()
        y_score = out.cpu().numpy().tolist()

        fpr, tpr, threshold = metrics.roc_curve(y_true, y_score)
        roc_auc = metrics.auc(fpr, tpr)

        aupr = metrics.average_precision_score(y_true, y_score)
        torch.cuda.empty_cache()

    return roc_auc, aupr

class SparseRowIndexer:
    def __init__(self, csr_matrix):
        data = []
        indices = []
        indptr = []

        for row_start, row_end in zip(csr_matrix.indptr[:-1], csr_matrix.indptr[1:]):
            data.append(csr_matrix.data[row_start:row_end])
            indices.append(csr_matrix.indices[row_start:row_end])
            indptr.append(row_end - row_start)  # nnz of the row

        self.data = np.array(data, dtype=object)
        self.indices = np.array(indices, dtype=object)
        self.indptr = np.array(indptr, dtype=object)
        self.shape = csr_matrix.shape

    def __getitem__(self, row_selector):
        indices = np.concatenate(self.indices[row_selector])
        data = np.concatenate(self.data[row_selector])
        indptr = np.append(0, np.cumsum(self.indptr[row_selector]))
        shape = [indptr.shape[0] - 1, self.shape[1]]
        return ssp.csr_matrix((data, indices, indptr), shape=shape)


class SparseColIndexer:
    def __init__(self, csc_matrix):
        data = []
        indices = []
        indptr = []

        for col_start, col_end in zip(csc_matrix.indptr[:-1], csc_matrix.indptr[1:]):
            data.append(csc_matrix.data[col_start:col_end])
            indices.append(csc_matrix.indices[col_start:col_end])
            indptr.append(col_end - col_start)

        self.data = np.array(data, dtype=object)
        self.indices = np.array(indices, dtype=object)
        self.indptr = np.array(indptr, dtype=object)
        self.shape = csc_matrix.shape

    def __getitem__(self, col_selector):
        indices = np.concatenate(self.indices[col_selector])
        data = np.concatenate(self.data[col_selector])
        indptr = np.append(0, np.cumsum(self.indptr[col_selector]))

        shape = [self.shape[0], indptr.shape[0] - 1]
        return ssp.csc_matrix((data, indices, indptr), shape=shape)


class MyDataset(InMemoryDataset):
    def __init__(self, root, A, links, labels, hop):
        self.Arow = SparseRowIndexer(A)
        self.Acol = SparseColIndexer(A.tocsc())
        self.links = links
        self.labels = labels
        self.hop = hop
        super(MyDataset, self).__init__(root)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def processed_file_names(self):
        name = 'data.pt'
        return [name]

    def process(self):
        # Extract enclosing subgraphs and save to disk
        data_list = links2subgraphs(self.Arow, self.Acol, self.links, self.labels, self.hop)
        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[0])
        del data_list


class MyDynamicDataset(Dataset):
    def __init__(self, root, A, links, labels, h):
        super(MyDynamicDataset, self).__init__(root)
        self.Arow = SparseRowIndexer(A)
        self.Acol = SparseColIndexer(A.tocsc())
        self.links = links
        self.labels = labels
        self.h = h

    def __len__(self):
        return len(self.links[0])

    def get(self, idx):
        i, j = self.links[0][idx], self.links[1][idx]
        g_label = self.labels[idx]
        tmp = subgraph_extraction_labeling(
            (i, j), self.Arow, self.Acol, g_label, self.h)
        return construct_pyg_graph(*tmp[0:6])


def links2subgraphs(Arow, Acol, links, labels, hop):
    # extract enclosing subgraphs
    print('Enclosing subgraph extraction begins...')
    g_list = []
    # with tqdm(total=len(links[0])) as pbar:
    #     num = 0
    #     for i, j, g_label in zip(links[0], links[1], labels):
    #         tmp = subgraph_extraction_labeling(
    #             (i, j), Arow, Acol, g_label, hop
    #         )
    #         data = construct_pyg_graph(*tmp)
    #         g_list.append(data)
    #         pbar.update(1)
    #         num = num + tmp[-1]

    start = time.time()
    pool = mp.Pool(mp.cpu_count())
    results = pool.starmap_async(
        subgraph_extraction_labeling,
        [
            ((i, j), Arow, Acol, g_label)
            for i, j, g_label in zip(links[0], links[1], labels)
        ]
    )
    remaining = results._number_left
    pbar = tqdm(total=remaining)
    while True:
        pbar.update(remaining - results._number_left)
        if results.ready(): break
        remaining = results._number_left
        time.sleep(1)
    results = results.get()
    pool.close()
    pbar.close()
    end = time.time()
    print("Time elapsed for subgraph extraction: {}s".format(end - start))
    print("Transforming to pytorch_geometric graphs...")
    g_list = []
    pbar = tqdm(total=len(results))
    while results:
        tmp = results.pop()
        g_list.append(construct_pyg_graph(*tmp[0:6]))
        pbar.update(1)
    pbar.close()
    end2 = time.time()
    print("Time elapsed for transforming to pytorch_geometric graphs: {}s".format(end2 - end))
            
    return g_list


def subgraph_extraction_labeling(ind, Arow, Acol, label=1, h=1):

    u_nodes, v_nodes = [ind[0]], [ind[1]]
    u_dist, v_dist = [0], [0]
    u_visited, v_visited = set([ind[0]]), set([ind[1]])
    u_fringe, v_fringe = set([ind[0]]), set([ind[1]])

    for dist in range(1, h+1):
        if len(u_fringe) == 0 or len(v_fringe) == 0:
            break

        v_fringe, u_fringe = neighbors(u_fringe, Arow), neighbors(v_fringe, Acol)
        u_fringe = u_fringe - u_visited
        v_fringe = v_fringe - v_visited
       
        u_visited = u_visited.union(u_fringe)
        v_visited = v_visited.union(v_fringe)

        u_nodes = u_nodes + list(u_fringe)
        v_nodes = v_nodes + list(v_fringe)
        u_dist = u_dist + [dist] * len(u_fringe)
        v_dist = v_dist + [dist] * len(v_fringe)

    subgraph = Arow[u_nodes][:, v_nodes]
    # remove link between target nodes
    subgraph[0, 0] = 0
    # prepare pyg graph constructor input
    u, v, r = ssp.find(subgraph)
    v += len(u_nodes)
    # num_nodes = len(u_nodes) + len(v_nodes)
    node_labels = [x*2 for x in u_dist] + [x*2+1 for x in v_dist]
    max_node_label = 2*8 + 1  # to construct initialize label trick matrix

    return u, v, r, node_labels, max_node_label, label


def construct_pyg_graph(u, v, r, node_labels, max_node_label, y):
    u, v = torch.LongTensor(u), torch.LongTensor(v)
    r = torch.LongTensor(r)  
    edge_index = torch.stack([torch.cat([u, v]), torch.cat([v, u])], 0)
    edge_type = torch.cat([r, r])
    x = torch.FloatTensor(one_hot(node_labels, max_node_label+1))
    y = torch.FloatTensor([y])
    data = Data(x, edge_index, edge_attr=edge_type, y=y)

    return data


def neighbors(fringe, A):
    # find all 1-hop neighbors of nodes in fringe from A
    return set(A[list(fringe)].indices)


def one_hot(idx, length):
    idx = np.array(idx)
    x = np.zeros([len(idx), length])
    x[np.arange(len(idx)), idx] = 1.0
    return x


def PyGGraph_to_nx(data):
    edges = list(zip(data.edge_index[0, :].tolist(), data.edge_index[1, :].tolist()))
    g = nx.from_edgelist(edges)
    # in case some nodes are isolated
    g.add_nodes_from(range(len(data.x)))
    edge_types = {(u, v): data.edge_type[i].item() for i, (u, v) in enumerate(edges)}
    nx.set_edge_attributes(g, name='type', values=edge_types)
    node_types = dict(zip(range(data.num_nodes), torch.argmax(data.x, 1).tolist()))
    nx.set_node_attributes(g, name='type', values=node_types)
    g.graph['rating'] = data.y.item()
    return g


#def load_k_fold(data_name, seed):
#    root_path = os.path.dirname(os.path.abspath(__file__))
#    if data_name == 'lrssl':
#        # txt dataset
#        path = os.path.join(root_path, 'drug_data/{}'.format(data_name) + '.txt')
#        matrix = pd.read_table(path, index_col=0).values
#    elif data_name in ['Gdataset', 'Cdataset']:
#        path = os.path.join(root_path, 'drug_data/{}'.format(data_name) + '.mat')
#        # mat dataset
#        data = io.loadmat(path)
#        matrix = data['didr'].T
#    else:
#        # csv dataset
#        path = os.path.join(root_path, 'drug_data/{}'.format(data_name) + '.csv')
#        data = pd.read_csv(path, header=None)
#        matrix = data.values.T

def load_k_fold(matrix, seed):

    drug_num, disease_num = matrix.shape[0], matrix.shape[1]
    drug_id, disease_id = np.nonzero(matrix)

    num_len = int(np.ceil(len(drug_id) * 1))  # setting sparse ratio
    drug_id, disease_id = drug_id[0: num_len], disease_id[0: num_len]

    neutral_flag = 0
    labels = np.full((drug_num, disease_num), neutral_flag, dtype=np.int32)
    observed_labels = [1] * len(drug_id)
    labels[drug_id, disease_id] = np.array(observed_labels)
    labels = labels.reshape([-1])

    # number of test and validation edges
    #num_train = int(np.ceil(0.9 * len(drug_id)))
    #num_test = int(np.ceil(0.1 * len(drug_id)))
    #print("num_train {}".format(num_train),
    #      "num_test {}".format(num_test))

    #print("num_train, num_test's ratio is", 0.9, 0.1)

    # negative sampling
    neg_drug_idx, neg_disease_idx = np.where(matrix == 0)
    neg_pairs = np.array([[dr, di] for dr, di in zip(neg_drug_idx, neg_disease_idx)])
    np.random.seed(6)
    np.random.shuffle(neg_pairs)
    # neg_pairs = neg_pairs[0:num_train + num_test - 1]
    neg_idx = np.array([dr * disease_num + di for dr, di in neg_pairs])

    # positive sampling
    pos_pairs = np.array([[dr, di] for dr, di in zip(drug_id, disease_id)])
    pos_idx = np.array([dr * disease_num + di for dr, di in pos_pairs])

    split_data_dict = {}
    count = 0
    #kfold = KFold(n_splits=k, shuffle=True, random_state=seed)
    #for train_data, test_data in kfold.split(pos_idx):
    for train_data in [pos_idx]:
        # train dataset contains positive and negative
        idx_pos_train = np.array(pos_idx)#[np.array(train_data)]

        idx_neg_train = neg_idx[0:len(idx_pos_train)]  # training dataset pos:neg = 1:1
        idx_train = np.concatenate([idx_pos_train, idx_neg_train], axis=0)

        pairs_pos_train = pos_pairs#[np.array(train_data)]
        pairs_neg_train = neg_pairs[0:len(pairs_pos_train)]
        pairs_train = np.concatenate([pairs_pos_train, pairs_neg_train], axis=0)

        # test dataset contains positive and negative
        #idx_pos_test = np.array(pos_idx)[np.array(test_data)]
        #idx_neg_test = neg_idx[len(pairs_pos_train): len(pairs_pos_train) + len(idx_pos_test) + 1]
        #idx_test = np.concatenate([idx_pos_test, idx_neg_test], axis=0)

        #pairs_pos_test = pos_pairs[np.array(test_data)]
        #pairs_neg_test = neg_pairs[len(pairs_pos_train): len(pairs_pos_train) + len(idx_pos_test) + 1]
        #pairs_test = np.concatenate([pairs_pos_test, pairs_neg_test], axis=0)

        # Internally shuffle training set
        rand_idx = list(range(len(idx_train)))
        np.random.seed(42)
        np.random.shuffle(rand_idx)
        idx_train = idx_train[rand_idx]
        pairs_train = pairs_train[rand_idx]

        u_train_idx, v_train_idx = pairs_train.transpose()
        #u_test_idx, v_test_idx = pairs_test.transpose()

        # create labels
        train_labels = labels[idx_train]
        #test_labels = labels[idx_test]

        # make training adjacency matrix
        rating_mx_train = np.zeros(drug_num * disease_num, dtype=np.float32)
        rating_mx_train[idx_train] = labels[idx_train]
        rating_mx_train = ssp.csr_matrix(rating_mx_train.reshape(drug_num, disease_num))
        split_data_dict[count] = [rating_mx_train, train_labels, u_train_idx, v_train_idx]#, \
        #   test_labels, u_test_idx, v_test_idx]
        count += 1

    return split_data_dict



class PSGCN(torch.nn.Module):
    def __init__(self, dataset, gconv=GCNConv, latent_dim=[64, 64, 1], k=30,
                 dropout=0.3, force_undirected=False):
        super(PSGCN, self).__init__()

        self.dropout = dropout
        self.force_undirected = force_undirected
        self.score1 = attention_score(latent_dim[0])
        self.score2 = attention_score(latent_dim[1])
        self.score3 = attention_score(latent_dim[2])

        self.conv1 = gconv(dataset.num_features, latent_dim[0])
        self.conv2 = gconv(latent_dim[0], latent_dim[1])
        self.conv3 = gconv(latent_dim[1], latent_dim[2])

        if k < 1:  # transform percentile to number
            node_nums = sorted([g.num_nodes for g in dataset])
            k = node_nums[int(math.ceil(k * len(node_nums)))-1]
            k = max(10, k)  # no smaller than 10

        self.k = int(k)
        self.dropout = dropout
        conv1d_channels = [16, 32]
        self.total_latent_dim = sum(latent_dim)
        conv1d_kws = [self.total_latent_dim, 5]
        self.conv1d_params1 = Conv1d(1, conv1d_channels[0], conv1d_kws[0], conv1d_kws[0])
        self.maxpool1d = nn.MaxPool1d(2, 2)

        self.conv1d_params2 = Conv1d(conv1d_channels[0], conv1d_channels[1], conv1d_kws[1], 1)
        dense_dim = int((k - 2) / 2 + 1)
        if (dense_dim<conv1d_kws[1]):
            dense_dim = conv1d_kws[1]
        self.dense_dim = (dense_dim - conv1d_kws[1] + 1) * conv1d_channels[1]
        self.lin1 = Linear(self.dense_dim, 128)
        self.lin2 = Linear(128, 1)

    def reset_parameters(self):
        self.conv1.reset_parameters()
        self.conv2.reset_parameters()
        self.conv3.reset_parameters()
        self.conv1d_params1.reset_parameters()
        self.conv1d_params2.reset_parameters()
        self.lin1.reset_parameters()
        self.lin2.reset_parameters()

    def forward(self, data):
        x, edge_index, batch = data.x, data.edge_index, data.batch

        # delete edge_attribute
        edge_index, edge_type = dropout_adj(
            edge_index, p=self.dropout,
            force_undirected=self.force_undirected, num_nodes=len(x),
            training=self.training
        )

        x = torch.relu(self.conv1(x, edge_index))
        attention_score1 = self.score1(x, edge_index)
        x1 = torch.mul(attention_score1, x)

        x = torch.relu(self.conv2(x, edge_index))
        attention_score2 = self.score2(x, edge_index)
        x2 = torch.mul(attention_score2, x)

        x = torch.relu((self.conv3(x, edge_index)))
        attention_score3 = self.score3(x, edge_index)
        x3 = torch.mul(attention_score3, x)

        X = [x1, x2, x3]
        concat_states = torch.cat(X, 1)

        x = global_sort_pool(concat_states, batch, self.k)
        x = x.unsqueeze(1)
        x = F.relu(self.conv1d_params1(x))
        x = self.maxpool1d(x)
        x = F.relu(self.conv1d_params2(x))
        x = x.view(len(x), -1)

        x = F.relu(self.lin1(x))
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.lin2(x)

        return x[:, 0]

class attention_score(torch.nn.Module):
    def __init__(self, in_channels, Conv=GCNConv):
        super(attention_score, self).__init__()
        self.in_channels = in_channels
        self.score_layer = Conv(in_channels, 1)

    def forward(self, x, edge_index, edge_attr=None, batch=None):
        score = self.score_layer(x, edge_index)

        return score


#coding: utf-8

## https://github.com/LiangXujun/LRSSL/tree/a16a75c028393e7256e3630bc8b7900026061f99
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler

import numpy as np
from subprocess import call
from functools import reduce

import calendar
import time
current_GMT = time.gmtime()

class LRSSL(BasicModel):
    def __init__(self, params=None):
        call("R -q -e \"print('R is installed and running.')\"", shell=True)
        self.LRSSL_filepath="LRSSL.R"
        cmd = "wget -qO - \'https://raw.githubusercontent.com/LiangXujun/LRSSL/a16a75c028393e7256e3630bc8b7900026061f99/LRSSL.R\' | sed -n \'/###/q;p\' > %s" % self.LRSSL_filepath
        call(cmd, shell=True)
        params = params if (params is not None) else self.default_parameters()
        super(LRSSL, self).__init__(params)
        self.scalerS, self.scalerP = None, None
        self.name = "LRSSL"
        self.estimator = None

    def default_parameters(self):
        params = {
            "k": 10, "mu": 0.01, "lam": 0.01, "gam": 2, "tol": 1e-2, "maxiter": 500, "sep_feature": "-",
        }
        return params

    def preprocessing(self, dataset, is_training=True, inf=2):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        S_ = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        S_ = np.nan_to_num(S_, nan=0.0) ##
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        P_ = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        P_ = np.nan_to_num(P_, nan=0.0) ##
        if (all([self.sep_feature in str(f) for f in dataset.item_features])):
            types_feature = [str(f).split(self.sep_feature)[0] for f in dataset.item_features]
            X_lst = [S_[:,np.argwhere(np.array(types_feature)==tf)].T for tf in list(set(types_feature))]
            X_lst = [x.reshape(x.shape[1:]) for x in X_lst]
        else:
            X_lst = [S_.T]
        X_p = P_ if (P_.shape[0]==P_.shape[1]) else np.corrcoef(P_)
        S_lst = [x.T.dot(x)/np.sqrt(x.sum(axis=0).dot(x.sum(axis=0).T)) for x in X_lst]
        S_lst = [s-np.diag(np.diag(s)) for s in S_lst]
        Y = dataset.ratings.toarray().copy()
        ST = np.zeros((Y.shape[0], Y.shape[0]))
        for i in range(Y.shape[0]):
            for j in range(i+1,Y.shape[0]):
                s = X_p[Y[i,:]==1,:][:,Y[j,:]==1]
                if (s.shape[1]==0 or s.shape[0]==0):
                    continue
                ST[i,j] = np.max(s)
        ST = ST + ST.T
        S_lst.append(ST)
        X_lst = [x.T if (x.shape[0]==x.shape[1]) else np.corrcoef(x.T) for x in X_lst]
        return [X_lst, S_lst, Y] if (is_training) else [X_lst, dataset.folds.shape]
        
    def model_fit(self, X_lst, S_lst, Y):
        time_stamp = calendar.timegm(current_GMT)
        filefolder = "LRSSL_%s/" % time_stamp
        call("mkdir -p %s/" % filefolder, shell=True)
        L_lst = []
        for s in S_lst:
            np.savetxt("%s/s.csv" % filefolder,s)
            call("R -q -e 'source(\"%s\");S <- as.matrix(read.table(\"%s/s.csv\", sep=\" \", header=F));Sp <- get.knn.graph(S, %d);write.csv(S, \"%s/s.csv\", row.names=F)' 2>&1 >/dev/null" % (self.LRSSL_filepath, filefolder, self.k, filefolder), shell=True)
            L_lst.append(np.loadtxt("%s/s.csv" % filefolder, skiprows=1, delimiter=","))
        L_lst = [np.diag(L.sum(axis=0))-L for L in L_lst]
        for i, x in enumerate(X_lst):
            np.savetxt("%s/x_%d.csv" % (filefolder, i+1),x)
        for i, l in enumerate(L_lst):
            np.savetxt("%s/l_%d.csv" % (filefolder, i+1),l)
        np.savetxt("%s/y.csv" % filefolder, Y)
        assert all([s.shape[0]==Y.shape[0] and s.shape[1]==Y.shape[0] for s in S_lst])
        assert all([l.shape[0]==Y.shape[0] and l.shape[1]==Y.shape[0] for l in L_lst])
        assert all([x.shape[0]==Y.shape[0] and x.shape[1]==Y.shape[0] for x in X_lst])
        call("R -q -e 'source(\"%s\");ml <- %d;mx <- %d;X_lst <- lapply(1:mx, function(i) as.matrix(read.table(paste0(\"%s/x_\",i,\".csv\"), sep=\" \", header=F)));L_lst <- lapply(1:ml, function(i) as.matrix(read.table(paste0(\"%s/l_\",i,\".csv\"), sep=\" \", header=F)));Y <- as.matrix(read.table(\"%s/y.csv\", sep=\" \", header=F));train.res <- lrssl(X_lst, L_lst, Y, mx, ml, %f, %f, %f, %d, %f);for(i in 1:mx){write.csv(train.res$Gs[[i]], paste0(\"%s/G_\",i,\".csv\"), row.names=F)};write.csv(train.res$alpha, \"%s/alpha.csv\", row.names=F);write.csv(list(t=train.res$t,diff_G=train.res$diff.G), \"%s/vals.csv\", row.names=F);write.csv(train.res$F.mat, \"%s/F_mat.csv\", row.names=F)' | grep '\[1\]'" % (self.LRSSL_filepath, len(L_lst), len(X_lst), filefolder, filefolder, filefolder, self.mu, self.lam, self.gam, self.maxiter, self.tol, filefolder, filefolder, filefolder, filefolder), shell=True)
        self.estimator = {
            "G": [np.loadtxt("%s/G_%d.csv" % (filefolder, i+1), skiprows=1, delimiter=",") for i in range(len(X_lst))],
            "alpha": np.loadtxt("%s/alpha.csv" % filefolder, skiprows=1, delimiter=","),
            "t": np.loadtxt("%s/vals.csv" % filefolder, skiprows=1, delimiter=",")[0],
            "diff_G": np.loadtxt("%s/vals.csv" % filefolder, skiprows=1, delimiter=",")[1],
            "F_mat": np.loadtxt("%s/F_mat.csv" % filefolder, skiprows=1, delimiter=","),
        }
        call("rm -rf %s/ %s" % (filefolder, self.LRSSL_filepath), shell=True)

    def model_predict_proba(self, X_lst, shp):
        Y_lst = [x.dot(self.estimator["G"][ix]) for ix, x in enumerate(X_lst)]
        Y_lst = [self.estimator["alpha"][iy]*(y/np.tile(y.sum(axis=1), (y.shape[1],1)).T) for iy, y in enumerate(Y_lst)]
        preds = np.ravel(reduce(sum, Y_lst))
        return preds.reshape(shp)


#coding: utf-8

from stanscofi.models import BasicModel
from fastai.collab import CollabDataLoaders, collab_learner
import numpy as np
import pandas as pd

#https://towardsdatascience.com/collaborative-filtering-using-fastai-a2ec5a2a4049
#https://docs.fast.ai/tutorial.collab.html#Interpretation
class FastaiCollabWrapper(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        name = "FastaiCollabWrapper"
        super(FastaiCollabWrapper, self).__init__(params)
        self.y_range = [-1.0,1.0]
        self.name = "FastaiCollabWrapper"

    def default_parameters(self):
        params = {
            "n_iterations": 5, 
            "n_factors" :20,
            "weight_decay" : 0.1,
            "learning_rate" : 5e-3,
            "random_state": 1354,
        }
        return params

    def preprocessing(self, dataset, is_training=True):
        ## only consider known ratings
        if (is_training):
            mat = np.column_stack((dataset.ratings.col, dataset.ratings.row, dataset.ratings.data))
            keep_ids = None
        ## consider all ratings
        else:
            ids = np.argwhere(np.ones(dataset.ratings.shape))
            mat = np.column_stack((ids[:,1].ravel(), ids[:,0].ravel(), dataset.ratings.toarray().ravel()))
            keep_ids = dataset.folds.toarray().ravel()!=0
        df = pd.DataFrame(mat, index=range(mat.shape[0]), columns=["disease","drug","rating"]).astype(int)
        return [df] if (is_training) else [df, keep_ids, ids.shape[0]]
    
    def model_fit(self, df):
        np.random.seed(self.random_state)
        data = CollabDataLoaders.from_df(df, user_name='disease', item_name="drug", rating_name="rating", bs=64, seed=self.random_state, valid_pct=0.05)
        self.model = collab_learner(data, n_factors=self.n_factors, use_nn=True, y_range=self.y_range, emb_szs=None)
        self.model.fit_one_cycle(self.n_iterations, self.learning_rate, wd=self.weight_decay)
    
    def model_predict_proba(self, df, keep_ids, n):
        ## https://docs.fast.ai/tutorial.tabular
        dl = self.model.dls.test_dl(df)
        preds = self.model.get_preds(dl=dl)[0].numpy().flatten()
        return preds[keep_ids]


import numpy as np
import torch
import dgl
import random
import seaborn
import os
from sklearn.metrics import roc_curve, roc_auc_score, \
    precision_recall_curve, average_precision_score
import matplotlib.pyplot as plt

## https://raw.githubusercontent.com/gu-yaowen/MilGNet/a0a129311a07feb928580ee7695d4321f24fd5c3/baseline/model_zoo.py
import dgl
import torch
import math
import torch.nn as nn
import dgl.function as fn
import torch.nn.functional as F
import dgl.nn as dglnn
from dgl.nn.pytorch import GATConv
from dgl.nn.functional import edge_softmax


class PGCN(nn.Module):

    def __init__(self, in_feats, rel_names):
        super().__init__()
        self.drug_emb = nn.Linear(in_feats[0], 64, bias=False)
        self.dis_emb = nn.Linear(in_feats[1], 64, bias=False)
        HeteroGraphdict_1 = {}
        for rel in rel_names:
            graphconv = dglnn.GraphConv(64, 64)
            nn.init.xavier_normal_(graphconv.weight)
            HeteroGraphdict_1[rel] = graphconv
        self.embedding_1 = dglnn.HeteroGraphConv(HeteroGraphdict_1, aggregate='sum')
        HeteroGraphdict_2 = {}
        for rel in rel_names:
            graphconv = dglnn.GraphConv(64, 32)
            nn.init.xavier_normal_(graphconv.weight)
            HeteroGraphdict_2[rel] = graphconv
        self.embedding_2 = dglnn.HeteroGraphConv(HeteroGraphdict_2, aggregate='sum')
        self.dropout = nn.Dropout(p=0.4)
        self.relu = nn.ReLU()
        self.weights = nn.Linear(32, 32, bias=False)
        nn.init.xavier_uniform_(self.weights.weight)

    def forward(self, g, feature):
        h = {'drug': self.drug_emb(feature['drug']),
             'disease': self.dis_emb(feature['disease'])}

        h = self.embedding_1(g, h)
        h = {k: self.relu(self.dropout(v)) for k, v in h.items()}

        h = self.embedding_2(g, h)
        h = {k: self.relu(self.dropout(v)) for k, v in h.items()}
        outputs = torch.matmul(self.weights(h['drug']), h['disease'].T)
        return outputs


class HeteroRGCNLayer(nn.Module):
    """Base heterodrugous RGCN layer.
    """

    def __init__(self, etypes, in_feats, out_feats):
        super(HeteroRGCNLayer, self).__init__()
        # W_r for each relation
        self.weight = nn.ModuleDict({
            name: nn.Linear(in_feats, out_feats) for name in etypes
        })

    def forward(self, g, feat_dict):
        # The input is a dictionary of node features for each type
        funcs = {}
        for srctype, etype, dsttype in g.canonical_etypes:
            # 计算每一类etype的 W_r * h
            Wh = self.weight[etype](feat_dict[srctype])
            # Save it in graph for message passing
            g.nodes[srctype].data['Wh_%s' % etype] = Wh
            # 消息函数 copy_u: 将源节点的特征聚合到'm'中; reduce函数: 将'm'求均值赋值给 'h'
            funcs[etype] = (fn.copy_u('Wh_%s' % etype, 'm'), fn.mean('m', 'h'))
        # Trigger message passing of multiple types.
        # The first argument is the message passing functions for each relation.
        # The second one is the type wise reducer, could be "sum", "max",
        # "min", "mean", "stack"
        g.multi_update_all(funcs, 'sum')
        # return the updated node feature dictionary
        return {ntype: g.nodes[ntype].data['h'] for ntype in g.ntypes}


class HeteroRGCN(nn.Module):
    def __init__(self, in_feats, rel_names):
        super(HeteroRGCN, self).__init__()
        self.drug_emb = nn.Linear(in_feats[0], 64, bias=False)
        self.dis_emb = nn.Linear(in_feats[1], 64, bias=False)
        self.embedding_1 = HeteroRGCNLayer(rel_names, 64, 32)
        self.embedding_2 = HeteroRGCNLayer(rel_names, 32, 32)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=0.4)
        self.weights = nn.Linear(32, 32, bias=False)
        nn.init.xavier_uniform_(self.weights.weight)

    def forward(self, g, feature):
        h = {'drug': self.drug_emb(feature['drug']),
             'disease': self.dis_emb(feature['disease'])}

        h = self.embedding_1(g, h)
        h = {k: self.relu(self.dropout(v)) for k, v in h.items()}

        h = self.embedding_2(g, h)
        h = {k: self.relu(self.dropout(v)) for k, v in h.items()}
        outputs = torch.matmul(self.weights(h['drug']), h['disease'].T)
        return outputs


class HGTLayer(nn.Module):
    def __init__(self,
                 in_dim,
                 out_dim,
                 node_dict,
                 edge_dict,
                 n_heads,
                 dropout=0.4,
                 use_norm=False):
        super(HGTLayer, self).__init__()

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.node_dict = node_dict
        self.edge_dict = edge_dict
        self.num_types = len(node_dict)
        self.num_relations = len(edge_dict)
        self.total_rel = self.num_types * self.num_relations * self.num_types
        self.n_heads = n_heads
        self.d_k = out_dim // n_heads
        self.sqrt_dk = math.sqrt(self.d_k)
        self.att = None

        self.k_linears = nn.ModuleList()
        self.q_linears = nn.ModuleList()
        self.v_linears = nn.ModuleList()
        self.a_linears = nn.ModuleList()
        self.norms = nn.ModuleList()
        self.use_norm = use_norm

        for t in range(self.num_types):
            self.k_linears.append(nn.Linear(in_dim, out_dim))
            self.q_linears.append(nn.Linear(in_dim, out_dim))
            self.v_linears.append(nn.Linear(in_dim, out_dim))
            self.a_linears.append(nn.Linear(out_dim, out_dim))
            if use_norm:
                self.norms.append(nn.LayerNorm(out_dim))

        self.relation_pri = nn.Parameter(torch.ones(self.num_relations, self.n_heads))
        self.relation_att = nn.Parameter(torch.Tensor(self.num_relations, n_heads, self.d_k, self.d_k))
        self.relation_msg = nn.Parameter(torch.Tensor(self.num_relations, n_heads, self.d_k, self.d_k))
        self.skip = nn.Parameter(torch.ones(self.num_types))
        self.drop = nn.Dropout(dropout)

        nn.init.xavier_uniform_(self.relation_att)
        nn.init.xavier_uniform_(self.relation_msg)

    def forward(self, G, h):
        with G.local_scope():
            node_dict, edge_dict = self.node_dict, self.edge_dict
            for srctype, etype, dsttype in G.canonical_etypes:
                sub_graph = G[srctype, etype, dsttype]

                k_linear = self.k_linears[node_dict[srctype]]
                v_linear = self.v_linears[node_dict[srctype]]
                q_linear = self.q_linears[node_dict[dsttype]]

                k = k_linear(h[srctype]).view(-1, self.n_heads, self.d_k)
                v = v_linear(h[srctype]).view(-1, self.n_heads, self.d_k)
                q = q_linear(h[dsttype]).view(-1, self.n_heads, self.d_k)

                e_id = self.edge_dict[etype]

                relation_att = self.relation_att[e_id]
                relation_pri = self.relation_pri[e_id]
                relation_msg = self.relation_msg[e_id]

                k = torch.einsum("bij,ijk->bik", k, relation_att)
                v = torch.einsum("bij,ijk->bik", v, relation_msg)

                sub_graph.srcdata['k'] = k
                sub_graph.dstdata['q'] = q
                sub_graph.srcdata['v_%d' % e_id] = v

                sub_graph.apply_edges(fn.v_dot_u('q', 'k', 't'))
                attn_score = sub_graph.edata.pop('t').sum(-1) * relation_pri / self.sqrt_dk
                attn_score = edge_softmax(sub_graph, attn_score, norm_by='dst')

                sub_graph.edata['t'] = attn_score.unsqueeze(-1)

            G.multi_update_all({etype: (fn.u_mul_e('v_%d' % e_id, 't', 'm'), fn.sum('m', 't')) \
                                for etype, e_id in edge_dict.items()}, cross_reducer='mean')

            new_h = {}
            for ntype in G.ntypes:
                '''
                    Step 3: Target-specific Aggregation
                    x = norm( W[node_type] * gelu( Agg(x) ) + x )
                '''
                n_id = node_dict[ntype]
                alpha = torch.sigmoid(self.skip[n_id])
                t = G.nodes[ntype].data['t'].view(-1, self.out_dim)
                trans_out = self.drop(self.a_linears[n_id](t))
                trans_out = trans_out * alpha + h[ntype] * (1 - alpha)
                if self.use_norm:
                    new_h[ntype] = self.norms[n_id](trans_out)
                else:
                    new_h[ntype] = trans_out
            return new_h


class HGT(nn.Module):
    def __init__(self, in_feats, node_dict, edge_dict):
        super(HGT, self).__init__()
        self.drug_emb = nn.Linear(in_feats[0], 64, bias=False)
        self.dis_emb = nn.Linear(in_feats[1], 64, bias=False)
        self.embedding = HGTLayer(64, 64, node_dict, edge_dict, 2, 0.1)
        self.relu = nn.ReLU()
        self.weights = nn.Linear(64, 64, bias=False)
        nn.init.xavier_uniform_(self.weights.weight)

    def forward(self, g, feature):
        h = {'drug': self.drug_emb(feature['drug']),
             'disease': self.dis_emb(feature['disease'])}

        h = self.embedding(g, h)
        h = {k: self.relu(v) for k, v in h.items()}

        outputs = torch.matmul(self.weights(h['drug']), h['disease'].T)

        return outputs


class SemanticAttention(nn.Module):
    def __init__(self, in_size, hidden_size=128):
        super(SemanticAttention, self).__init__()

        self.project = nn.Sequential(
            nn.Linear(in_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, 1, bias=False)
        )

    def forward(self, z):
        w = self.project(z).mean(0)  # (M, 1)
        beta = torch.softmax(w, dim=0)  # (M, 1)
        beta = beta.expand((z.shape[0],) + beta.shape)  # (N, M, 1)

        return (beta * z).sum(1)  # (N, D * K)


class HANLayer(nn.Module):
    """
    HAN layer.
    Arguments
    ---------
    meta_paths : list of metapaths, each as a list of edge types
    in_size : input feature dimension
    out_size : output feature dimension
    layer_num_heads : number of attention heads
    dropout : Dropout probability
    Inputs
    ------
    g : DGLHeteroGraph
        The heterodrugous graph
    h : tensor
        Input features
    Outputs
    -------
    tensor
        The output feature
    """

    def __init__(self, meta_paths, in_size, out_size, layer_num_heads, dropout):
        super(HANLayer, self).__init__()

        # One GAT layer for each meta path based adjacency matrix
        self.gat_layers = nn.ModuleList()
        for i in range(len(meta_paths)):
            self.gat_layers.append(GATConv(in_size, out_size, layer_num_heads,
                                           dropout, dropout, activation=F.elu,
                                           allow_zero_in_degree=True))
        self.semantic_attention = SemanticAttention(in_size=out_size * layer_num_heads)
        self.meta_paths = list(tuple(meta_path) for meta_path in meta_paths)

        self._cached_graph = None
        self._cached_coalesced_graph = {}

    def forward(self, g, h):
        semantic_embeddings = []

        if self._cached_graph is None or self._cached_graph is not g:
            self._cached_graph = g
            self._cached_coalesced_graph.clear()
            for meta_path in self.meta_paths:
                self._cached_coalesced_graph[meta_path] = dgl.metapath_reachable_graph(
                    g, meta_path)

        for i, meta_path in enumerate(self.meta_paths):
            new_g = self._cached_coalesced_graph[meta_path]
            semantic_embeddings.append(self.gat_layers[i](new_g, h).flatten(1))
        semantic_embeddings = torch.stack(semantic_embeddings, dim=1)  # (N, M, D * K)

        return self.semantic_attention(semantic_embeddings)  # (N, D * K)


class HAN(nn.Module):
    def __init__(self, meta_paths, in_feats):
        super(HAN, self).__init__()
        self.drug_emb = nn.Linear(in_feats[0], 64, bias=False)
        self.dis_emb = nn.Linear(in_feats[1], 64, bias=False)
        self.embedding_1_drug = HANLayer([meta_paths[0]], 64, 32, 5, 0.4)
        self.embedding_1_dis = HANLayer([meta_paths[1]], 64, 32, 5, 0.4)
        self.drug_emb_1 = nn.Linear(32 * 5, 32, bias=False)
        self.dis_emb_1 = nn.Linear(32 * 5, 32, bias=False)
        self.embedding_2_drug = HANLayer([meta_paths[0]], 32, 32, 5, 0.4)
        self.embedding_2_dis = HANLayer([meta_paths[1]], 32, 32, 5, 0.4)
        self.drug_emb_2 = nn.Linear(32 * 5, 32, bias=False)
        self.dis_emb_2 = nn.Linear(32 * 5, 32, bias=False)
        self.weights = nn.Linear(32, 32, bias=False)
        nn.init.xavier_uniform_(self.weights.weight)

    def forward(self, g, feature):
        h = {'drug': self.drug_emb(feature['drug']),
             'disease': self.dis_emb(feature['disease'])}

        h['drug'] = self.drug_emb_1(self.embedding_1_drug(g, h['drug']))
        h['disease'] = self.dis_emb_1(self.embedding_1_dis(g, h['disease']))
        h['drug'] = self.drug_emb_2(self.embedding_2_drug(g, h['drug']))
        h['disease'] = self.dis_emb_2(self.embedding_2_dis(g, h['disease']))
        outputs = torch.matmul(self.weights(h['drug']), h['disease'].T)

        return outputs


## https://raw.githubusercontent.com/gu-yaowen/MilGNet/a0a129311a07feb928580ee7695d4321f24fd5c3/baseline/utils.py
def topk_filtering(d_d: np.array, k: int):
    """Convert the Topk similarities to 1 and generate the Topk interactions.
    """
    for i in range(len(d_d)):
        sorted_idx = np.argpartition(d_d[i], -k - 1)
        d_d[i, sorted_idx[-k - 1:-1]] = 1
    return np.array(np.where(d_d == 1)).T


def remove_graph(g, test_drug_id, test_disease_id):
    etype = ('drug', 'drug-disease', 'disease')

    edges_id = g.edge_ids(torch.tensor(test_drug_id),
                          torch.tensor(test_disease_id),
                          etype=etype)
    g = dgl.remove_edges(g, edges_id, etype=etype)
    etype = ('disease', 'disease-drug', 'drug')
    edges_id = g.edge_ids(torch.tensor(test_disease_id),
                          torch.tensor(test_drug_id),
                          etype=etype)
    g = dgl.remove_edges(g, edges_id, etype=etype)
    return g


def get_metrics_auc(real_score, predict_score):
    AUC = roc_auc_score(real_score, predict_score)
    AUPR = average_precision_score(real_score, predict_score)
    return AUC, AUPR


def get_metrics(real_score, predict_score):
    """Calculate the performance metrics.
    Resource code is acquired from:
    Yu Z, Huang F, Zhao X et al.
     Predicting drug-disease associations through layer attention graph convolutional network,
     Brief Bioinform 2021;22.

    Parameters
    ----------
    real_score: true labels
    predict_score: model predictions

    Return
    ---------
    AUC, AUPR, Accuracy, F1-Score, Precision, Recall, Specificity
    """
    sorted_predict_score = np.array(
        sorted(list(set(np.array(predict_score).flatten()))))
    sorted_predict_score_num = len(sorted_predict_score)
    thresholds = sorted_predict_score[np.int32(
        sorted_predict_score_num * np.arange(1, 1000) / 1000)]
    thresholds = np.mat(thresholds)
    thresholds_num = thresholds.shape[1]

    predict_score_matrix = np.tile(predict_score, (thresholds_num, 1))
    negative_index = np.where(predict_score_matrix < thresholds.T)
    positive_index = np.where(predict_score_matrix >= thresholds.T)
    predict_score_matrix[negative_index] = 0
    predict_score_matrix[positive_index] = 1
    TP = predict_score_matrix.dot(real_score.T)
    FP = predict_score_matrix.sum(axis=1) - TP
    FN = real_score.sum() - TP
    TN = len(real_score.T) - TP - FP - FN

    fpr = FP / (FP + TN)
    tpr = TP / (TP + FN)
    ROC_dot_matrix = np.mat(sorted(np.column_stack((fpr, tpr)).tolist())).T
    ROC_dot_matrix.T[0] = [0, 0]
    ROC_dot_matrix = np.c_[ROC_dot_matrix, [1, 1]]
    x_ROC = ROC_dot_matrix[0].T
    y_ROC = ROC_dot_matrix[1].T
    auc = 0.5 * (x_ROC[1:] - x_ROC[:-1]).T * (y_ROC[:-1] + y_ROC[1:])

    recall_list = tpr
    precision_list = TP / (TP + FP)
    PR_dot_matrix = np.mat(sorted(np.column_stack(
        (recall_list, precision_list)).tolist())).T
    PR_dot_matrix.T[0] = [0, 1]
    PR_dot_matrix = np.c_[PR_dot_matrix, [1, 0]]
    x_PR = PR_dot_matrix[0].T
    y_PR = PR_dot_matrix[1].T
    aupr = 0.5 * (x_PR[1:] - x_PR[:-1]).T * (y_PR[:-1] + y_PR[1:])

    f1_score_list = 2 * TP / (len(real_score.T) + TP - TN)
    accuracy_list = (TP + TN) / len(real_score.T)
    specificity_list = TN / (TN + FP)

    max_index = np.argmax(f1_score_list)
    f1_score = f1_score_list[max_index]
    accuracy = accuracy_list[max_index]
    specificity = specificity_list[max_index]
    recall = recall_list[max_index]
    precision = precision_list[max_index]
    return auc[0, 0], aupr[0, 0], accuracy, f1_score, precision, recall, specificity


def set_seed(seed=0):
    print('seed = {}'.format(seed))
    os.environ['PYTHONHASHSEED'] = str(seed)
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['MKL_NUM_THREADS'] = '1'
    torch.set_num_threads(1)
    random.seed(seed)
    np.random.seed(seed)
    # dgl.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.use_deterministic_algorithms(True)
    torch.backends.cudnn.enabled = False
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


def plot_result_auc(args, label, predict, auc):
    """Plot the ROC curve for predictions.
    Parameters
    ----------
    args: argumentation
    label: true labels
    predict: model predictions
    auc: calculated AUROC score
    """
    seaborn.set_style()
    fpr, tpr, threshold = roc_curve(label, predict)
    plt.figure()
    lw = 2
    plt.figure(figsize=(8, 8))
    plt.plot(fpr, tpr, color='darkorange',
             lw=lw, label='ROC curve (area = %0.4f)' % auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic example')
    plt.legend(loc='lower right')
    plt.savefig(os.path.join(args.saved_path, 'result_auc.png'))
    plt.clf()


def plot_result_aupr(args, label, predict, aupr):
    """Plot the ROC curve for predictions.
    Parameters
    ----------
    args: argumentation
    label: true labels
    predict: model predictions
    aupr: calculated AUPR score
    """
    seaborn.set_style()
    precision, recall, thresholds = precision_recall_curve(label, predict)
    plt.figure()
    lw = 2
    plt.figure(figsize=(8, 8))
    plt.plot(precision, recall, color='darkorange',
             lw=lw, label='AUPR Score (area = %0.4f)' % aupr)
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.title('RPrecision/Recall Curve')
    plt.legend(loc='lower right')
    plt.savefig(os.path.join(args.saved_path, 'result_aupr.png'))
    plt.clf()



#coding: utf-8

## https://github.com/luckymengmeng/SCPMF/tree/9158aa22d53687938bf6402dbcd9812aabf557e9
from stanscofi.models import BasicModel
from stanscofi.preprocessing import CustomScaler

import numpy as np
import os
from subprocess import call

import calendar
import time
current_GMT = time.gmtime()

## /!\ Only tested on Linux
class SCPMF(BasicModel):
    def __init__(self, params=None):
        try:
            call("octave -v", shell=True)
        except:
            raise ValueError("Please install Octave.")
        params = params if (params is not None) else self.default_parameters()
        super(SCPMF, self).__init__(params)
        self.scalerS, self.scalerP = None, None
        self.name = "SCPMF" 
        self.predictions = None
        self.SCPMF_filepath = None

    def default_parameters(self):
        params = {
            "r": 15,
        }
        return params

    def preprocessing(self, dataset, is_training=True, inf=2):
        if (self.scalerS is None):
            self.scalerS = CustomScaler(posinf=inf, neginf=-inf)
        S_ = self.scalerS.fit_transform(dataset.items.T.toarray().copy(), subset=None)
        X_s = S_ if (S_.shape[0]==S_.shape[1]) else np.corrcoef(S_)
        if (self.scalerP is None):
            self.scalerP = CustomScaler(posinf=inf, neginf=-inf)
        P_ = self.scalerP.fit_transform(dataset.users.T.toarray().copy(), subset=None)
        X_p = P_ if (P_.shape[0]==P_.shape[1]) else np.corrcoef(P_)
        A_sp = np.copy(dataset.ratings.toarray()) # items x users
        return [X_s, X_p, A_sp]
        
    def model_fit(self, X_s, X_p, A_sp):
        time_stamp = calendar.timegm(current_GMT)
        filefolder = "SCPMF_%s" % time_stamp 
        call("mkdir -p %s/" % filefolder, shell=True)
        call("wget -qO %s/SCPMFDR.m 'https://raw.githubusercontent.com/luckymengmeng/SCPMF/master/SCPMFDR.m'" % filefolder, shell=True)
        np.savetxt("%s/X_p.csv" % filefolder, X_p, delimiter=",")
        np.savetxt("%s/A_sp.csv" % filefolder, A_sp, delimiter=",")
        np.savetxt("%s/X_s.csv" % filefolder, X_s, delimiter=",")
        cmd = "A_sp = csvread('A_sp.csv');X_s = csvread('X_s.csv');X_p = csvread('X_p.csv');recMatrix=SCPMFDR(A_sp,X_s,X_p,%d);csvwrite('recMatrix.csv', recMatrix);" % (self.r)
        call("cd %s/ && octave --silent --eval \"%s\"" % (filefolder, cmd), shell=True)
        self.predictions = np.loadtxt("%s/recMatrix.csv" % filefolder, delimiter=",")
        call("rm -rf %s/" % filefolder, shell=True)
  
    def model_predict_proba(self, X_s, X_p, A_sp):
        return self.predictions


#coding: utf-8

from stanscofi.models import BasicModel
import numpy as np
from scipy.sparse import csr_matrix

from benchscofi.implementations import BayesianPairwiseRanking

#' Matrix factorization
class PMF(BasicModel):
    def __init__(self, params=None):
        params = params if (params is not None) else self.default_parameters()
        super(PMF, self).__init__(params)
        self.name = "PMF"
        self.estimator = BayesianPairwiseRanking.BPR(**params)

    def default_parameters(self):
        params = BayesianPairwiseRanking.bpr_params
        return params

    def preprocessing(self, dataset, is_training=True):
        ## Use 0-1 ratings (matrix form)
        return [csr_matrix(dataset.ratings.toarray().T)]
        
    def model_fit(self, X):
        self.estimator.fit(X)

    def model_predict_proba(self, X):
        preds = self.estimator.predict().T
        return preds
